from jambo.exceptions.internal_assertion_exception import InternalAssertionException

from unittest import TestCase


class TestInternalAssertionException(TestCase):
    def test_inheritance(self):
        self.assertTrue(issubclass(InternalAssertionException, RuntimeError))

    def test_message(self):
        message = "This is an internal assertion error."

        expected_message = (
            f"Internal Assertion Failed: {message}\n"
            "This is likely a bug in Jambo. Please report it at"
        )

        with self.assertRaises(InternalAssertionException) as ctx:
            raise InternalAssertionException(message)

        self.assertEqual(str(ctx.exception), expected_message)



from jambo.exceptions import InternalAssertionException, InvalidSchemaException
from jambo.parser import GenericTypeParser
from jambo.types import RefCacheDict
from jambo.types.json_schema_type import JSONSchema
from jambo.types.type_parser_options import TypeParserOptions

from typing_extensions import ForwardRef, Literal, Union, Unpack


RefType = Union[type, ForwardRef]

RefStrategy = Literal["forward_ref", "def_ref"]


class RefTypeParser(GenericTypeParser):
    json_schema_type = "$ref"

    def from_properties_impl(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ) -> tuple[RefType, dict]:
        if "$ref" not in properties:
            raise InvalidSchemaException(
                f"Missing $ref in properties for {name}", invalid_field="$ref"
            )

        if kwargs.get("context") is None:
            raise InternalAssertionException(
                "`context` must be provided in kwargs for RefTypeParser"
            )

        ref_cache = kwargs.get("ref_cache")
        if ref_cache is None:
            raise InternalAssertionException(
                "`ref_cache` must be provided in kwargs for RefTypeParser"
            )

        mapped_properties = self.mappings_properties_builder(properties, **kwargs)

        ref_strategy, ref_name, ref_property = self._examine_ref_strategy(
            name, properties, **kwargs
        )

        ref_state = self._get_ref_from_cache(ref_name, ref_cache)
        if ref_state is not None:
            # If the reference is either processing or already cached
            return ref_state, mapped_properties

        ref = self._parse_from_strategy(ref_strategy, ref_name, ref_property, **kwargs)
        ref_cache[ref_name] = ref

        return ref, mapped_properties

    def _parse_from_strategy(
        self,
        ref_strategy: RefStrategy,
        ref_name: str,
        ref_property: JSONSchema,
        **kwargs: Unpack[TypeParserOptions],
    ) -> RefType:
        mapped_type: RefType
        match ref_strategy:
            case "forward_ref":
                mapped_type = ForwardRef(ref_name)
            case "def_ref":
                mapped_type, _ = GenericTypeParser.type_from_properties(
                    ref_name, ref_property, **kwargs
                )
            case _:
                raise InvalidSchemaException(
                    f"Unsupported $ref {ref_property['$ref']}", invalid_field="$ref"
                )

        return mapped_type

    def _get_ref_from_cache(
        self, ref_name: str, ref_cache: RefCacheDict
    ) -> RefType | type | None:
        try:
            ref_state = ref_cache[ref_name]

            if ref_state is None:
                # If the reference is being processed, we return a ForwardRef
                return ForwardRef(ref_name)

            # If the reference is already cached, we return it
            return ref_state
        except KeyError:
            # If the reference is not in the cache, we will set it to None
            ref_cache[ref_name] = None

        return None

    def _examine_ref_strategy(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ) -> tuple[RefStrategy, str, JSONSchema]:
        if properties.get("$ref") == "#":
            ref_name = kwargs["context"].get("title")
            if ref_name is None:
                raise InvalidSchemaException(
                    "Missing title in properties for $ref of Root Reference",
                    invalid_field="title",
                )
            return "forward_ref", ref_name, {}

        if properties.get("$ref", "").startswith("#/$defs/"):
            target_name, target_property = self._extract_target_ref(
                name, properties, **kwargs
            )
            return "def_ref", target_name, target_property

        raise InvalidSchemaException(
            "Only Root and $defs references are supported at the moment",
            invalid_field="$ref",
        )

    def _extract_target_ref(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ) -> tuple[str, JSONSchema]:
        target_name = None
        target_property = kwargs["context"]
        for prop_name in properties["$ref"].split("/")[1:]:
            if prop_name not in target_property:
                raise InvalidSchemaException(
                    f"Missing {prop_name} in properties for $ref {properties['$ref']}",
                    invalid_field=prop_name,
                )
            target_name = prop_name
            target_property = target_property[prop_name]  # type: ignore

        if not isinstance(target_name, str) or target_property is None:
            raise InvalidSchemaException(
                f"Invalid $ref {properties['$ref']}", invalid_field="$ref"
            )

        return target_name, target_property



from jambo.exceptions import InternalAssertionException
from jambo.parser import ObjectTypeParser

from unittest import TestCase


class TestObjectTypeParser(TestCase):
    def test_object_type_parser_throws_without_ref_cache(self):
        parser = ObjectTypeParser()

        properties = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
        }

        with self.assertRaises(InternalAssertionException):
            parser.from_properties_impl("placeholder", properties)

    def test_object_type_parser(self):
        parser = ObjectTypeParser()

        properties = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
        }

        Model, _args = parser.from_properties_impl(
            "placeholder", properties, ref_cache={}
        )

        obj = Model(name="name", age=10)

        self.assertEqual(obj.name, "name")
        self.assertEqual(obj.age, 10)

    def test_object_type_parser_with_object_example(self):
        parser = ObjectTypeParser()

        properties = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "examples": [
                {
                    "name": "example_name",
                    "age": 30,
                }
            ],
        }

        _, type_validator = parser.from_properties_impl(
            "placeholder", properties, ref_cache={}
        )

        test_example = type_validator["examples"][0]

        self.assertEqual(test_example.name, "example_name")
        self.assertEqual(test_example.age, 30)

    def test_object_type_parser_with_default(self):
        parser = ObjectTypeParser()

        properties = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "default": {
                "name": "default_name",
                "age": 20,
            },
        }

        _, type_validator = parser.from_properties_impl(
            "placeholder", properties, ref_cache={}
        )

        # Check default value
        default_obj = type_validator["default_factory"]()
        self.assertEqual(default_obj.name, "default_name")
        self.assertEqual(default_obj.age, 20)

        # Chekc default factory new object id
        new_obj = type_validator["default_factory"]()
        self.assertNotEqual(id(default_obj), id(new_obj))

    def test_object_type_parser_warns_if_object_override_in_cache(self):
        ref_cache = {}

        parser = ObjectTypeParser()

        properties = {"type": "object", "properties": {}}

        with self.assertWarns(UserWarning):
            _, type_validator = parser.from_properties_impl(
                "placeholder", properties, ref_cache=ref_cache
            )
            _, type_validator = parser.from_properties_impl(
                "placeholder", properties, ref_cache=ref_cache
            )



from jambo.parser._type_parser import GenericTypeParser
from jambo.types.type_parser_options import TypeParserOptions

from typing_extensions import Unpack


class NullTypeParser(GenericTypeParser):
    mapped_type = type(None)

    json_schema_type = "type:null"

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        mapped_properties = self.mappings_properties_builder(properties, **kwargs)
        mapped_properties["default"] = None

        return self.mapped_type, mapped_properties



from jambo.parser._type_parser import GenericTypeParser
from jambo.types.type_parser_options import TypeParserOptions

from typing_extensions import Unpack


class IntTypeParser(GenericTypeParser):
    mapped_type = int

    json_schema_type = "type:integer"

    type_mappings = {
        "minimum": "ge",
        "exclusiveMinimum": "gt",
        "maximum": "le",
        "exclusiveMaximum": "lt",
        "multipleOf": "multiple_of",
        "default": "default",
    }

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        return int, self.mappings_properties_builder(properties, **kwargs)



from jambo.exceptions import InvalidSchemaException, UnsupportedSchemaException
from jambo.parser import ObjectTypeParser, RefTypeParser
from jambo.types import JSONSchema, RefCacheDict

from jsonschema.exceptions import SchemaError
from jsonschema.validators import validator_for
from pydantic import BaseModel
from typing_extensions import Optional


class SchemaConverter:
    """
    Converts JSON Schema to Pydantic models.

    This class is responsible for converting JSON Schema definitions into Pydantic models.
    It validates the schema and generates the corresponding Pydantic model with appropriate
    fields and types. The generated model can be used for data validation and serialization.
    """

    def __init__(
        self, namespace_registry: Optional[dict[str, RefCacheDict]] = None
    ) -> None:
        if namespace_registry is None:
            namespace_registry = dict()
        self._namespace_registry = namespace_registry

    def build_with_cache(
        self,
        schema: JSONSchema,
        ref_cache: Optional[RefCacheDict] = None,
        without_cache: bool = False,
    ) -> type[BaseModel]:
        """
        Converts a JSON Schema to a Pydantic model.
        This is the instance method version of `build` and uses the instance's reference cache if none is provided.
        Use this method if you want to utilize the instance's reference cache.

            :param schema: The JSON Schema to convert.
            :param ref_cache: An optional reference cache to use during conversion.
            :param without_cache: Whether to use a clean reference cache for this conversion.
            :return: The generated Pydantic model.
        """
        local_ref_cache: RefCacheDict

        if without_cache:
            local_ref_cache = dict()
        elif ref_cache is None:
            namespace = schema.get("$id", "default")
            local_ref_cache = self._namespace_registry.setdefault(namespace, dict())
        else:
            local_ref_cache = ref_cache

        return self.build(schema, local_ref_cache)

    @staticmethod
    def build(
        schema: JSONSchema, ref_cache: Optional[RefCacheDict] = None
    ) -> type[BaseModel]:
        """
        Converts a JSON Schema to a Pydantic model.
        This method doesn't use a reference cache if none is provided.
            :param schema: The JSON Schema to convert.
            :param ref_cache: An optional reference cache to use during conversion, if provided `with_clean_cache` will be ignored.
            :return: The generated Pydantic model.
        """
        if ref_cache is None:
            ref_cache = dict()

        try:
            validator = validator_for(schema)
            validator.check_schema(schema)  # type: ignore
        except SchemaError as err:
            raise InvalidSchemaException(
                "Validation of JSON Schema failed.", cause=err
            ) from err

        if "title" not in schema:
            raise InvalidSchemaException(
                "Schema must have a title.", invalid_field="title"
            )

        schema_type = SchemaConverter._get_schema_type(schema)

        match schema_type:
            case "object":
                return ObjectTypeParser.to_model(
                    schema["title"],
                    schema.get("properties", {}),
                    schema.get("required", []),
                    context=schema,
                    ref_cache=ref_cache,
                    required=True,
                )

            case "$ref":
                parsed_model, _ = RefTypeParser().from_properties(
                    schema["title"],
                    schema,
                    context=schema,
                    ref_cache=ref_cache,
                    required=True,
                )
                return parsed_model
            case _:
                unsupported_type = (
                    f"type:{schema_type}" if schema_type else "missing type"
                )
                raise UnsupportedSchemaException(
                    "Only object and $ref schema types are supported.",
                    unsupported_field=unsupported_type,
                )

    def clear_ref_cache(self, namespace: Optional[str] = "default") -> None:
        """
        Clears the reference cache.
        """
        if namespace is None:
            self._namespace_registry.clear()
            return

        if namespace in self._namespace_registry:
            self._namespace_registry[namespace].clear()

    def get_cached_ref(
        self, ref_name: str, namespace: str = "default"
    ) -> Optional[type]:
        """
        Gets a cached reference from the reference cache.
        :param ref_name: The name of the reference to get.
        :return: The cached reference, or None if not found.
        """        
        cached_type = self._namespace_registry.get(
            namespace, {}
        ).get(ref_name)

        if isinstance(cached_type, type):
            return cached_type

        return None

    @staticmethod
    def _get_schema_type(schema: JSONSchema) -> str | None:
        """
        Returns the type of the schema.
        :param schema: The JSON Schema to check.
        :return: The type of the schema.
        """
        if "$ref" in schema:
            return "$ref"

        type_value = schema.get("type")
        if isinstance(type_value, list):
            raise InvalidSchemaException(
                "Invalid schema: 'type' cannot be a list at the top level",
                invalid_field=str(schema),
            )

        return type_value



class InternalAssertionException(RuntimeError):
    """Exception raised for internal assertions."""

    def __init__(
        self,
        message: str,
    ) -> None:
        # Normalize message by stripping redundant prefix if present
        message = message.removeprefix("Internal Assertion Failed: ")
        super().__init__(message)

    def __str__(self) -> str:
        return (
            f"Internal Assertion Failed: {super().__str__()}\n"
            "This is likely a bug in Jambo. Please report it at"
        )



from jambo.exceptions import InternalAssertionException, InvalidSchemaException
from jambo.parser import ObjectTypeParser, RefTypeParser

from pydantic import ValidationError
from typing_extensions import ForwardRef

from unittest import TestCase


class TestRefTypeParser(TestCase):
    def test_ref_type_parser_throws_without_ref(self):
        properties = {
            "title": "person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "required": ["name", "age"],
        }

        with self.assertRaises(InvalidSchemaException):
            RefTypeParser().from_properties(
                "person",
                properties,
                context=properties,
                ref_cache={},
                required=True,
            )

    def test_ref_type_parser_throws_without_context(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
        }

        with self.assertRaises(InternalAssertionException):
            RefTypeParser().from_properties(
                "person",
                properties,
                ref_cache={},
                required=True,
            )

    def test_ref_type_parser_throws_without_ref_cache(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
        }

        with self.assertRaises(InternalAssertionException):
            RefTypeParser().from_properties(
                "person",
                properties,
                context=properties,
                required=True,
            )

    def test_ref_type_parser_throws_if_network_ref_type(self):
        properties = {
            "title": "person",
            "$ref": "https://example.com/schemas/person.json",
        }

        with self.assertRaises(InvalidSchemaException):
            RefTypeParser().from_properties(
                "person",
                properties,
                context=properties,
                ref_cache={},
                required=True,
            )

    def test_ref_type_parser_throws_if_non_root_or_def_ref(self):
        # This is invalid because object3 is referencing object2,
        # but object2 is not defined in $defs or as a root reference.
        properties = {
            "title": "object1",
            "type": "object",
            "properties": {
                "object2": {
                    "type": "object",
                    "properties": {
                        "attr1": {
                            "type": "string",
                        },
                        "attr2": {
                            "type": "integer",
                        },
                    },
                },
                "object3": {
                    "$ref": "#/$defs/object2",
                },
            },
        }

        with self.assertRaises(InvalidSchemaException):
            ObjectTypeParser().from_properties(
                "person",
                properties,
                context=properties,
                ref_cache={},
                required=True,
            )

    def test_ref_type_parser_throws_if_def_doesnt_exists(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/employee",
            "$defs": {},
        }

        with self.assertRaises(InvalidSchemaException):
            RefTypeParser().from_properties(
                "person",
                properties,
                context=properties,
                ref_cache={},
                required=True,
            )

    def test_ref_type_parser_throws_if_ref_property_doesnt_exists(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {"person": None},
        }

        with self.assertRaises(InvalidSchemaException):
            RefTypeParser().from_properties(
                "person",
                properties,
                context=properties,
                ref_cache={},
                required=True,
            )

    def test_ref_type_parser_with_def(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
        }

        type_parsing, type_validator = RefTypeParser().from_properties(
            "person",
            properties,
            context=properties,
            ref_cache={},
            required=True,
        )

        self.assertIsInstance(type_parsing, type)

        obj = type_parsing(name="John", age=30)

        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)

    def test_ref_type_parser_with_forward_ref(self):
        properties = {
            "title": "person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model, type_validator = ObjectTypeParser().from_properties(
            "person",
            properties,
            context=properties,
            ref_cache={},
            required=True,
        )

        obj = model(
            name="John",
            age=30,
            emergency_contact=model(
                name="Jane",
                age=28,
            ),
        )

        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)
        self.assertIsInstance(obj.emergency_contact, model)
        self.assertEqual(obj.emergency_contact.name, "Jane")
        self.assertEqual(obj.emergency_contact.age, 28)

    def test_ref_type_parser_invalid_forward_ref(self):
        properties = {
            # Doesn't have a title, which is required for forward references
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        with self.assertRaises(InvalidSchemaException):
            ObjectTypeParser().from_properties(
                "person",
                properties,
                context=properties,
                ref_cache={},
                required=True,
            )

    def test_ref_type_parser_forward_ref_can_checks_validation(self):
        properties = {
            "title": "person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model, type_validator = ObjectTypeParser().from_properties(
            "person",
            properties,
            context=properties,
            ref_cache={},
            required=True,
        )

        # checks if when created via FowardRef the model is validated correctly.
        with self.assertRaises(ValidationError):
            model(
                name="John",
                age=30,
                emergency_contact=model(
                    name="Jane",
                ),
            )

    def test_ref_type_parser_with_ciclic_def(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                        "emergency_contact": {
                            "$ref": "#/$defs/person",
                        },
                    },
                }
            },
        }

        model, type_validator = RefTypeParser().from_properties(
            "person",
            properties,
            context=properties,
            ref_cache={},
            required=True,
        )

        obj = model(
            name="John",
            age=30,
            emergency_contact=model(
                name="Jane",
                age=28,
            ),
        )

        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)
        self.assertIsInstance(obj.emergency_contact, model)
        self.assertEqual(obj.emergency_contact.name, "Jane")
        self.assertEqual(obj.emergency_contact.age, 28)

    def test_ref_type_parser_with_repeated_ref(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                        "emergency_contact": {
                            "$ref": "#/$defs/person",
                        },
                        "friends": {
                            "type": "array",
                            "items": {
                                "$ref": "#/$defs/person",
                            },
                        },
                    },
                }
            },
        }

        model, type_validator = RefTypeParser().from_properties(
            "person",
            properties,
            context=properties,
            ref_cache={},
            required=True,
        )

        obj = model(
            name="John",
            age=30,
            emergency_contact=model(
                name="Jane",
                age=28,
            ),
            friends=[
                model(name="Alice", age=25),
                model(name="Bob", age=26),
            ],
        )

        self.assertEqual(
            type(obj.emergency_contact),
            type(obj.friends[0]),
            "Emergency contact and friends should be of the same type",
        )

    def test_ref_type_parser_pre_computed_ref_cache(self):
        ref_cache = {}

        parent_properties = {
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
        }

        properties1 = {
            "title": "person1",
            "$ref": "#/$defs/person",
        }
        model1, _ = RefTypeParser().from_properties(
            "person",
            properties1,
            context=parent_properties,
            ref_cache=ref_cache,
            required=True,
        )

        properties2 = {
            "title": "person2",
            "$ref": "#/$defs/person",
        }
        model2, _ = RefTypeParser().from_properties(
            "person",
            properties2,
            context=parent_properties,
            ref_cache=ref_cache,
            required=True,
        )

        self.assertIs(model1, model2, "Models should be the same instance")

    def test_parse_from_strategy_invalid_ref_strategy(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
        }

        with self.assertRaises(InvalidSchemaException):
            ref_strategy, ref_name, ref_property = RefTypeParser()._parse_from_strategy(
                "invalid_strategy",
                "person",
                properties,
            )

    def test_parse_from_strategy_forward_ref(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
        }

        parsed_type = RefTypeParser()._parse_from_strategy(
            "forward_ref",
            "person",
            properties,
        )

        self.assertIsInstance(parsed_type, ForwardRef)

    def test_parse_from_strategy_def_ref(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
        }

        parsed_type = RefTypeParser()._parse_from_strategy(
            "def_ref",
            "person",
            properties,
            context=properties,
            ref_cache={},
            required=True,
        )

        obj = parsed_type(
            name="John",
            age=30,
        )

        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)

    def test_ref_type_parser_with_def_with_examples(self):
        properties = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                    },
                }
            },
            "examples": [
                {"name": "John", "age": 30},
                {"name": "Jane", "age": 25},
            ],
        }

        _, type_validator = RefTypeParser().from_properties(
            "person",
            properties,
            context=properties,
            ref_cache={},
            required=True,
        )

        self.assertEqual(
            type_validator.get("examples"),
            [
                {"name": "John", "age": 30},
                {"name": "Jane", "age": 25},
            ],
        )



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.type_parser_options import TypeParserOptions

from pydantic import BaseModel, BeforeValidator, Field, TypeAdapter, ValidationError
from typing_extensions import Annotated, Any, Union, Unpack, get_args


Annotation = Annotated[Any, ...]


class OneOfTypeParser(GenericTypeParser):
    mapped_type = Union

    json_schema_type = "oneOf"

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        if "oneOf" not in properties:
            raise InvalidSchemaException(
                f"Invalid JSON Schema: {properties}", invalid_field="oneOf"
            )

        if not isinstance(properties["oneOf"], list) or len(properties["oneOf"]) == 0:
            raise InvalidSchemaException(
                f"Invalid JSON Schema: {properties['oneOf']}", invalid_field="oneOf"
            )

        mapped_properties = self.mappings_properties_builder(properties, **kwargs)

        sub_types = [
            GenericTypeParser.type_from_properties(
                f"{name}_sub{i}", subProperty, **kwargs
            )
            for i, subProperty in enumerate(properties["oneOf"])
        ]

        if not kwargs.get("required", False):
            mapped_properties["default"] = mapped_properties.get("default")

        subfield_types = [Annotated[t, Field(**v)] for t, v in sub_types]

        # Added with the understanding of discriminator are not in the JsonSchema Spec,
        # they were added by OpenAPI and not all implementations may support them,
        # and they do not always generate a model one-to-one to the Pydantic model
        # TL;DR: Discriminators were added by OpenAPI and not a Official JSON Schema feature
        if (discriminator := properties.get("discriminator")) is not None:
            validated_type = self._build_type_one_of_with_discriminator(
                subfield_types, discriminator
            )
        else:
            validated_type = self._build_type_one_of_with_func(subfield_types)

        return validated_type, mapped_properties

    @staticmethod
    def _build_type_one_of_with_discriminator(
        subfield_types: list[Annotation], discriminator_prop: dict
    ) -> Annotation:
        """
        Build a type with a discriminator.
        """
        if not isinstance(discriminator_prop, dict):
            raise InvalidSchemaException(
                "Discriminator must be a dictionary", invalid_field="discriminator"
            )

        for field in subfield_types:
            field_type, field_info = get_args(field)

            if issubclass(field_type, BaseModel):
                continue

            raise InvalidSchemaException(
                "When using a discriminator, all subfield types must be of type 'object'.",
                invalid_field="discriminator",
            )

        property_name = discriminator_prop.get("propertyName")
        if property_name is None or not isinstance(property_name, str):
            raise InvalidSchemaException(
                "Discriminator must have a 'propertyName' key",
                invalid_field="propertyName",
            )

        return Annotated[Union[(*subfield_types,)], Field(discriminator=property_name)]

    @staticmethod
    def _build_type_one_of_with_func(subfield_types: list[Annotation]) -> Annotation:
        """
        Build a type with a validation function for the oneOf constraint.
        """

        def validate_one_of(value: Any) -> Any:
            matched_count = 0

            for field_type in subfield_types:
                try:
                    TypeAdapter(field_type).validate_python(value)
                    matched_count += 1
                except ValidationError:
                    continue

            if matched_count == 0:
                raise ValueError("Value does not match any of the oneOf schemas")
            elif matched_count > 1:
                raise ValueError(
                    "Value matches multiple oneOf schemas, exactly one expected"
                )

            return value

        return Annotated[Union[(*subfield_types,)], BeforeValidator(validate_one_of)]



from jambo import SchemaConverter
from jambo.exceptions import InvalidSchemaException, UnsupportedSchemaException
from jambo.types import JSONSchema

from pydantic import AnyUrl, BaseModel, ValidationError
from typing_extensions import get_args

from ipaddress import IPv4Address, IPv6Address
from unittest import TestCase
from uuid import UUID


def is_pydantic_model(cls):
    return isinstance(cls, type) and issubclass(cls, BaseModel)


class TestSchemaConverter(TestCase):
    def setUp(self):
        self.converter = SchemaConverter()

    def tearDown(self):
        self.converter.clear_ref_cache(namespace=None)

    def test_invalid_schema(self):
        schema = {
            "title": 1,
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
        }

        with self.assertRaises(InvalidSchemaException):
            self.converter.build_with_cache(schema)

    def test_invalid_schema_type(self):
        schema = {
            "title": 1,
            "description": "A person",
            "type": 1,
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
        }

        with self.assertRaises(InvalidSchemaException):
            self.converter.build_with_cache(schema)

    def test_build_expects_title(self):
        schema = {
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
        }

        with self.assertRaises(InvalidSchemaException):
            self.converter.build_with_cache(schema)

    def test_build_expects_object(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "string",
        }

        with self.assertRaises(UnsupportedSchemaException):
            self.converter.build_with_cache(schema)

    def test_is_invalid_field(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "id": {
                    "notType": "string",
                }
            },
            # 'required': ['name', 'age', 'is_active', 'friends', 'address'],
        }

        with self.assertRaises(InvalidSchemaException) as context:
            self.converter.build_with_cache(schema)
            self.assertTrue("Unknown type" in str(context.exception))

    def test_jsonschema_to_pydantic(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
            },
            "required": ["name"],
        }

        model = self.converter.build_with_cache(schema)

        self.assertTrue(is_pydantic_model(model))

    def test_validation_string(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {"type": "string", "maxLength": 4, "minLength": 1},
                "email": {
                    "type": "string",
                    "maxLength": 50,
                    "minLength": 5,
                    "pattern": r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$",
                },
            },
            "required": ["name"],
        }

        model = self.converter.build_with_cache(schema)

        self.assertEqual(model(name="John", age=30).name, "John")

        with self.assertRaises(ValidationError):
            model(name=123, age=30, email="teste@hideyoshi.com")

        with self.assertRaises(ValidationError):
            model(name="John Invalid", age=45, email="teste@hideyoshi.com")

        with self.assertRaises(ValidationError):
            model(name="", age=45, email="teste@hideyoshi.com")

        with self.assertRaises(ValidationError):
            model(name="John", age=45, email="hideyoshi.com")

    def test_validation_integer(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "age": {
                    "type": "integer",
                    "minimum": 0,
                    "maximum": 120,
                },
            },
            "required": ["age"],
        }

        model = self.converter.build_with_cache(schema)

        self.assertEqual(model(age=30).age, 30)

        with self.assertRaises(ValidationError):
            model(age=-1)

        with self.assertRaises(ValidationError):
            model(age=121)

    def test_validation_float(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "age": {
                    "type": "number",
                    "minimum": 0,
                    "maximum": 120,
                },
            },
            "required": ["age"],
        }

        model = self.converter.build_with_cache(schema)

        self.assertEqual(model(age=30).age, 30.0)

        with self.assertRaises(ValidationError):
            model(age=-1.0)

        with self.assertRaises(ValidationError):
            model(age=121.0)

    def test_validation_boolean(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "is_active": {"type": "boolean"},
            },
            "required": ["is_active"],
        }

        model = self.converter.build_with_cache(schema)

        self.assertEqual(model(is_active=True).is_active, True)

        self.assertEqual(model(is_active="true").is_active, True)

    def test_validation_list_with_valid_items(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "friends": {
                    "type": "array",
                    "items": {"type": "string"},
                    "minItems": 1,
                    "maxItems": 2,
                    "uniqueItems": True,
                },
            },
            "required": ["friends"],
        }

        model = self.converter.build_with_cache(schema)

        self.assertEqual(
            model(friends=["John", "Jane", "John"]).friends, {"John", "Jane"}
        )

        with self.assertRaises(ValidationError):
            model(friends=[])

        with self.assertRaises(ValidationError):
            model(friends=["John", "Jane", "Invalid"])

    def test_validation_list_with_missing_items(self):
        model = self.converter.build_with_cache(
            {
                "title": "Person",
                "description": "A person",
                "type": "object",
                "properties": {
                    "friends": {
                        "type": "array",
                        "items": {"type": "string"},
                        "minItems": 1,
                        "maxItems": 2,
                    },
                },
                "required": ["friends"],
            }
        )

        with self.assertRaises(ValidationError):
            model()

    def test_validation_object(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "required": ["street", "city"],
                },
            },
            "required": ["address"],
        }

        model = self.converter.build_with_cache(schema)

        obj = model(address={"street": "123 Main St", "city": "Springfield"})

        self.assertEqual(obj.address.street, "123 Main St")
        self.assertEqual(obj.address.city, "Springfield")

        with self.assertRaises(ValidationError):
            model()

    def test_default_for_string(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "default": "John",
                },
            },
            "required": ["name"],
        }

        model = self.converter.build_with_cache(schema)

        obj = model(name="John")

        self.assertEqual(obj.name, "John")

    def test_invalid_default_for_string(self):
        # Test for default with maxLength
        schema_max_length = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "default": "John",
                    "maxLength": 2,
                },
            },
            "required": ["name"],
        }

        with self.assertRaises(InvalidSchemaException):
            self.converter.build_with_cache(schema_max_length)

    def test_default_for_list(self):
        schema_list = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "friends": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["John", "Jane"],
                },
            },
            "required": ["friends"],
        }

        model_list = self.converter.build_with_cache(schema_list)

        self.assertEqual(model_list().friends, ["John", "Jane"])

    def test_default_for_list_with_unique_items(self):
        # Test for default with uniqueItems
        schema_set = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "friends": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["John", "Jane"],
                    "uniqueItems": True,
                },
            },
            "required": ["friends"],
        }

        model_set = self.converter.build_with_cache(schema_set)

        self.assertEqual(model_set().friends, {"John", "Jane"})

    def test_default_for_object(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "default": {"street": "123 Main St", "city": "Springfield"},
                },
            },
            "required": ["address"],
        }

        model = self.converter.build_with_cache(schema)

        obj = model(address={"street": "123 Main St", "city": "Springfield"})

        self.assertEqual(obj.address.street, "123 Main St")
        self.assertEqual(obj.address.city, "Springfield")

    def test_all_of(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "name": {
                    "allOf": [
                        {"type": "string", "maxLength": 11},
                        {"type": "string", "maxLength": 4},
                        {"type": "string", "minLength": 1},
                        {"type": "string", "minLength": 2},
                    ]
                },
            },
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model(
            name="J",
        )

        self.assertEqual(obj.name, "J")

        with self.assertRaises(ValidationError):
            Model(name="John Invalid")

        with self.assertRaises(ValidationError):
            Model(name="")

    def test_any_of(self):
        schema = {
            "title": "Person",
            "description": "A person",
            "type": "object",
            "properties": {
                "id": {
                    "anyOf": [
                        {"type": "string", "maxLength": 11, "minLength": 1},
                        {"type": "integer", "maximum": 10},
                    ]
                },
            },
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model(id=1)
        self.assertEqual(obj.id, 1)

        obj = Model(id="12345678901")
        self.assertEqual(obj.id, "12345678901")

        with self.assertRaises(ValidationError):
            Model(id="")

        with self.assertRaises(ValidationError):
            Model(id="12345678901234567890")

        with self.assertRaises(ValidationError):
            Model(id=11)

    def test_string_format_email(self):
        schema = {
            "title": "EmailTest",
            "type": "object",
            "properties": {"email": {"type": "string", "format": "email"}},
        }

        model = self.converter.build_with_cache(schema)
        self.assertEqual(model(email="test@example.com").email, "test@example.com")

        with self.assertRaises(ValidationError):
            model(email="invalid-email")

    def test_string_format_uri(self):
        schema = {
            "title": "UriTest",
            "type": "object",
            "properties": {"website": {"type": "string", "format": "uri"}},
        }

        model = self.converter.build_with_cache(schema)
        self.assertEqual(
            model(website="https://example.com").website, AnyUrl("https://example.com")
        )

        with self.assertRaises(ValidationError):
            model(website="invalid-uri")

    def test_string_format_ipv4(self):
        schema = {
            "title": "IPv4Test",
            "type": "object",
            "properties": {"ip": {"type": "string", "format": "ipv4"}},
        }

        model = self.converter.build_with_cache(schema)
        self.assertEqual(model(ip="192.168.1.1").ip, IPv4Address("192.168.1.1"))

        with self.assertRaises(ValidationError):
            model(ip="256.256.256.256")

    def test_string_format_ipv6(self):
        schema = {
            "title": "IPv6Test",
            "type": "object",
            "properties": {"ip": {"type": "string", "format": "ipv6"}},
        }

        model = self.converter.build_with_cache(schema)
        self.assertEqual(
            model(ip="2001:0db8:85a3:0000:0000:8a2e:0370:7334").ip,
            IPv6Address("2001:0db8:85a3:0000:0000:8a2e:0370:7334"),
        )

        with self.assertRaises(ValidationError):
            model(ip="invalid-ipv6")

    def test_string_format_uuid(self):
        schema = {
            "title": "UUIDTest",
            "type": "object",
            "properties": {"id": {"type": "string", "format": "uuid"}},
        }

        model = self.converter.build_with_cache(schema)

        self.assertEqual(
            model(id="123e4567-e89b-12d3-a456-426614174000").id,
            UUID("123e4567-e89b-12d3-a456-426614174000"),
        )

        with self.assertRaises(ValidationError):
            model(id="invalid-uuid")

    def test_string_format_hostname(self):
        schema = {
            "title": "HostnameTest",
            "type": "object",
            "properties": {"hostname": {"type": "string", "format": "hostname"}},
        }

        model = self.converter.build_with_cache(schema)
        self.assertEqual(model(hostname="example.com").hostname, "example.com")

        with self.assertRaises(ValidationError):
            model(hostname="invalid..hostname")

    def test_string_format_datetime(self):
        schema = {
            "title": "DateTimeTest",
            "type": "object",
            "properties": {"timestamp": {"type": "string", "format": "date-time"}},
        }

        model = self.converter.build_with_cache(schema)
        self.assertEqual(
            model(timestamp="2024-01-01T12:00:00Z").timestamp.isoformat(),
            "2024-01-01T12:00:00+00:00",
        )

        with self.assertRaises(ValidationError):
            model(timestamp="invalid-datetime")

    def test_string_format_time(self):
        schema = {
            "title": "TimeTest",
            "type": "object",
            "properties": {"time": {"type": "string", "format": "time"}},
        }

        model = self.converter.build_with_cache(schema)
        self.assertEqual(
            model(time="20:20:39+00:00").time.isoformat(), "20:20:39+00:00"
        )

        with self.assertRaises(ValidationError):
            model(time="25:00:00")

    def test_string_format_unsupported(self):
        schema = {
            "title": "InvalidFormat",
            "type": "object",
            "properties": {"field": {"type": "string", "format": "unsupported"}},
        }

        with self.assertRaises(InvalidSchemaException):
            self.converter.build_with_cache(schema)

    def test_ref_with_root_ref(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model = self.converter.build_with_cache(schema)

        obj = model(
            name="John",
            age=30,
            emergency_contact=model(
                name="Jane",
                age=28,
            ),
        )

        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)
        self.assertIsInstance(obj.emergency_contact, model)
        self.assertEqual(obj.emergency_contact.name, "Jane")
        self.assertEqual(obj.emergency_contact.age, 28)

    def test_ref_with_def(self):
        schema = {
            "title": "person",
            "$ref": "#/$defs/person",
            "$defs": {
                "person": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "integer"},
                        "emergency_contact": {
                            "$ref": "#/$defs/person",
                        },
                    },
                }
            },
        }

        model = self.converter.build_with_cache(schema)

        obj = model(
            name="John",
            age=30,
            emergency_contact=model(
                name="Jane",
                age=28,
            ),
        )

        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)
        self.assertIsInstance(obj.emergency_contact, model)
        self.assertEqual(obj.emergency_contact.name, "Jane")
        self.assertEqual(obj.emergency_contact.age, 28)

    def test_ref_with_def_another_model(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "address": {"$ref": "#/$defs/Address"},
            },
            "required": ["name"],
            "$defs": {
                "Address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "required": ["street", "city"],
                }
            },
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model(
            name="John",
            age=30,
            address={"street": "123 Main St", "city": "Springfield"},
        )

        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)
        self.assertEqual(obj.address.street, "123 Main St")
        self.assertEqual(obj.address.city, "Springfield")

    def test_enum_type_parser(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "status": {
                    "type": "string",
                    "enum": ["active", "inactive", "pending"],
                }
            },
            "required": ["status"],
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model(status="active")
        self.assertEqual(obj.status.value, "active")

    def test_enum_type_parser_with_default(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "status": {
                    "type": "string",
                    "enum": ["active", "inactive", "pending"],
                    "default": "active",
                }
            },
            "required": ["status"],
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model()
        self.assertEqual(obj.status.value, "active")

    def test_const_type_parser(self):
        schema = {
            "title": "Country",
            "type": "object",
            "properties": {
                "name": {
                    "const": "United States of America",
                }
            },
            "required": ["name"],
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model()
        self.assertEqual(obj.name, "United States of America")

        with self.assertRaises(ValidationError):
            obj.name = "Canada"

        with self.assertRaises(ValidationError):
            Model(name="Canada")

    def test_const_type_parser_with_non_hashable_value(self):
        schema = {
            "title": "Country",
            "type": "object",
            "properties": {
                "name": {
                    "const": ["Brazil"],
                }
            },
            "required": ["name"],
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model()
        self.assertEqual(obj.name, ["Brazil"])

        with self.assertRaises(ValidationError):
            obj.name = ["Argentina"]

        with self.assertRaises(ValidationError):
            Model(name=["Argentina"])

    def test_null_type_parser(self):
        schema = {
            "title": "Test",
            "type": "object",
            "properties": {
                "a_thing": {"type": "null"},
            },
        }

        Model = self.converter.build_with_cache(schema)

        obj = Model()
        self.assertIsNone(obj.a_thing)

        obj = Model(a_thing=None)
        self.assertIsNone(obj.a_thing)

        with self.assertRaises(ValidationError):
            Model(a_thing="not none")

    def test_scoped_ref_schema(self):
        schema: JSONSchema = {
            "title": "Example Schema",
            "type": "object",
            "properties": {
                "operating_system": {
                    "oneOf": [
                        {"$ref": "#/$defs/operating_system"},
                        {
                            "type": "object",
                            "properties": {
                                "creation": {"$ref": "#/$defs/operating_system"},
                                "reinstallation": {"$ref": "#/$defs/operating_system"},
                            },
                            "required": ["creation", "reinstallation"],
                        },
                    ]
                },
            },
            "$defs": {
                "operating_system": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "version": {"type": "string"},
                    },
                    "required": ["name", "version"],
                }
            },
        }

        schema_type = self.converter.build_with_cache(schema)

        # check for me that the types generated by the oneOf in the typing.Annotated have different names
        operating_system_field = schema_type.model_fields["operating_system"]

        arg1, arg2 = get_args(operating_system_field.annotation)

        first_type = get_args(arg1)[0]
        second_type = get_args(arg2)[0]

        self.assertNotEqual(first_type.__name__, second_type.__name__)

    def test_object_invalid_require(self):
        # https://github.com/HideyoshiNakazone/jambo/issues/60
        object_ = self.converter.build_with_cache(
            {
                "$schema": "https://json-schema.org/draft/2020-12/schema",
                "title": "TEST",
                "type": "object",
                "required": ["title"],
                "properties": {
                    "title": {
                        "type": "string",
                        "description": "The title of the object",
                    },
                    "description": {
                        "type": "object",
                        "properties": {
                            "summary": {
                                "type": "string",
                            },
                            "details": {
                                "type": "string",
                            },
                        },
                    },
                },
            }
        )

        self.assertFalse(object_.model_fields["description"].is_required())  # FAIL

    def test_instance_level_ref_cache(self):
        ref_cache = {}

        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        converter1 = SchemaConverter(ref_cache)
        model1 = converter1.build_with_cache(schema)

        converter2 = SchemaConverter(ref_cache)
        model2 = converter2.build_with_cache(schema)

        self.assertIs(model1, model2)

    def test_instance_level_ref_cache_isolation_via_without_cache_param(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model1 = self.converter.build_with_cache(schema, without_cache=True)
        model2 = self.converter.build_with_cache(schema, without_cache=True)

        self.assertIsNot(model1, model2)

    def test_instance_level_ref_cache_isolation_via_provided_cache(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model1 = self.converter.build_with_cache(schema, ref_cache={})
        model2 = self.converter.build_with_cache(schema, ref_cache={})

        self.assertIsNot(model1, model2)

    def test_get_type_from_cache(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model = self.converter.build_with_cache(schema)

        cached_model = self.converter.get_cached_ref("Person")

        self.assertIs(model, cached_model)

    def test_get_type_from_cache_not_found(self):
        cached_model = self.converter.get_cached_ref("NonExistentModel")

        self.assertIsNone(cached_model)

    def test_get_type_from_cache_nested_type(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "required": ["street", "city"],
                },
            },
            "required": ["name", "age", "address"],
        }

        model = self.converter.build_with_cache(schema)

        cached_model = self.converter.get_cached_ref("Person.address")

        self.assertIsNotNone(cached_model)
        self.assertIs(model.model_fields["address"].annotation, cached_model)

    def test_get_type_from_cache_with_def(self):
        schema = {
            "title": "person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "address": {"$ref": "#/$defs/address"},
            },
            "$defs": {
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "required": ["street", "city"],
                }
            },
        }

        person_model = self.converter.build_with_cache(schema)
        cached_person_model = self.converter.get_cached_ref("person")

        self.assertIs(person_model, cached_person_model)

        cached_address_model = self.converter.get_cached_ref("address")

        self.assertIsNotNone(cached_address_model)

    def test_parse_list_type_multiple_values(self):
        schema = {
            "title": "TestListType",
            "type": "object",
            "properties": {"values": {"type": ["string", "number"]}},
        }

        Model = self.converter.build_with_cache(schema)

        obj1 = Model(values="a string")
        self.assertEqual(obj1.values, "a string")

        obj2 = Model(values=42)
        self.assertEqual(obj2.values, 42)

    def test_parse_list_type_one_value(self):
        schema = {
            "title": "TestListType",
            "type": "object",
            "properties": {"values": {"type": ["string"]}},
        }

        Model = self.converter.build_with_cache(schema)

        obj1 = Model(values="a string")
        self.assertEqual(obj1.values, "a string")

    def test_parse_list_type_empty(self):
        schema = {
            "title": "TestListType",
            "type": "object",
            "properties": {"values": {"type": []}},
        }

        with self.assertRaises(InvalidSchemaException):
            self.converter.build_with_cache(schema)

    def test_parse_list_type_root_level_throws(self):
        schema = {"title": "TestListType", "type": ["string", "number"]}

        with self.assertRaises(InvalidSchemaException):
            self.converter.build_with_cache(schema)

    def tests_instance_level_ref_cache_isolation_via_property_id(self):
        schema1: JSONSchema = {
            "$id": "http://example.com/schemas/person1.json",
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model1 = self.converter.build_with_cache(schema1)

        schema2: JSONSchema = {
            "$id": "http://example.com/schemas/person2.json",
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "address": {"type": "string"},
            },
            "required": ["name", "age", "address"],
        }

        model2 = self.converter.build_with_cache(schema2)

        self.assertIsNot(model1, model2)

    def tests_instance_level_ref_cache_colision_when_same_property_id(self):
        schema1: JSONSchema = {
            "$id": "http://example.com/schemas/person.json",
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "emergency_contact": {
                    "$ref": "#",
                },
            },
            "required": ["name", "age"],
        }

        model1 = self.converter.build_with_cache(schema1)

        schema2: JSONSchema = {
            "$id": "http://example.com/schemas/person.json",
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "address": {"type": "string"},
            },
            "required": ["name", "age", "address"],
        }

        model2 = self.converter.build_with_cache(schema2)

        self.assertIs(model1, model2)

    def test_namespace_isolation_via_on_call_config(self):
        namespace = "namespace1"
        
        schema: JSONSchema = {
            "$id": namespace,
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "required": ["street", "city"],
                },
            },
            "required": ["name", "age", "address"],
        }

        model = self.converter.build_with_cache(schema)
        
        invalid_cached_model = self.converter.get_cached_ref("Person")
        self.assertIsNone(invalid_cached_model)
        
        cached_model = self.converter.get_cached_ref("Person", namespace=namespace)
        self.assertIs(model, cached_model)

    def test_clear_namespace_registry(self):
        namespace = "namespace_to_clear"
        
        schema: JSONSchema = {
            "$id": namespace,
            "title": "Person",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "integer"},
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                    },
                    "required": ["street", "city"],
                },
            },
            "required": ["name", "age", "address"],
        }

        model = self.converter.build_with_cache(schema)
        
        cached_model = self.converter.get_cached_ref("Person", namespace=namespace)
        self.assertIs(model, cached_model)
        
        self.converter.clear_ref_cache(namespace=namespace)
        
        cleared_cached_model = self.converter.get_cached_ref("Person", namespace=namespace)
        self.assertIsNone(cleared_cached_model)



from jambo.types.json_schema_type import JSONSchema

from typing_extensions import ForwardRef, MutableMapping, TypedDict


RefCacheDict = MutableMapping[str, ForwardRef | type | None]


class TypeParserOptions(TypedDict):
    required: bool
    context: JSONSchema
    ref_cache: RefCacheDict



from jambo.exceptions import InvalidSchemaException
from jambo.parser import BooleanTypeParser

from unittest import TestCase


class TestBoolTypeParser(TestCase):
    def test_bool_parser_no_options(self):
        parser = BooleanTypeParser()

        properties = {"type": "boolean"}

        type_parsing, type_validator = parser.from_properties_impl(
            "placeholder", properties
        )

        self.assertEqual(type_parsing, bool)
        self.assertEqual(type_validator, {"default": None})

    def test_bool_parser_with_default(self):
        parser = BooleanTypeParser()

        properties = {
            "type": "boolean",
            "default": True,
        }

        type_parsing, type_validator = parser.from_properties_impl(
            "placeholder", properties
        )

        self.assertEqual(type_parsing, bool)
        self.assertEqual(type_validator["default"], True)

    def test_bool_parser_with_invalid_default(self):
        parser = BooleanTypeParser()

        properties = {
            "type": "boolean",
            "default": "invalid",
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties_impl("placeholder", properties)

    def test_bool_parser_with_examples(self):
        parser = BooleanTypeParser()

        properties = {
            "type": "boolean",
            "examples": [True, False],
        }

        type_parsing, type_validator = parser.from_properties_impl(
            "placeholder", properties
        )

        self.assertEqual(type_parsing, bool)
        self.assertEqual(type_validator["default"], None)
        self.assertEqual(type_validator["examples"], [True, False])



from jambo.exceptions import InvalidSchemaException
from jambo.types.type_parser_options import JSONSchema, TypeParserOptions

from pydantic import Field, TypeAdapter
from typing_extensions import Annotated, Any, ClassVar, Generic, Self, TypeVar, Unpack

from abc import ABC, abstractmethod


T = TypeVar("T", bound=type)


class GenericTypeParser(ABC, Generic[T]):
    json_schema_type: ClassVar[str]

    type_mappings: dict[str, str] = {}

    default_mappings = {
        "default": "default",
        "description": "description",
        "examples": "examples",
        "title": "title",
        "deprecated": "deprecated",
    }

    @abstractmethod
    def from_properties_impl(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ) -> tuple[T, dict]:
        """
        Abstract method to convert properties to a type and its fields properties.
        :param name: The name of the type.
        :param properties: The properties of the type.
        :param kwargs: Additional options for type parsing.
        :return: A tuple containing the type and its properties.
        """

    def from_properties(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ) -> tuple[T, dict]:
        """
        Converts properties to a type and its fields properties.
        :param name: The name of the type.
        :param properties: The properties of the type.
        :param kwargs: Additional options for type parsing.
        :return: A tuple containing the type and its properties.
        """
        parsed_type, parsed_properties = self.from_properties_impl(
            name, properties, **kwargs
        )

        if not self._validate_default(parsed_type, parsed_properties):
            raise InvalidSchemaException(
                "Default value is not valid", invalid_field=name
            )

        if not self._validate_examples(parsed_type, parsed_properties):
            raise InvalidSchemaException(
                "Examples values are not valid", invalid_field=name
            )

        return parsed_type, parsed_properties

    @classmethod
    def type_from_properties(
        cls, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ) -> tuple[type, dict]:
        """
        Factory method to fetch the appropriate type parser based on properties
        and generates the equivalent type and fields.
        :param name: The name of the type to be created.
        :param properties: The properties that define the type.
        :param kwargs: Additional options for type parsing.
        :return: A tuple containing the type and its properties.
        """

        parser = cls._get_impl(cls._normalize_properties(properties))

        return parser().from_properties(name=name, properties=properties, **kwargs)

    @staticmethod
    def _normalize_properties(properties: JSONSchema) -> JSONSchema:
        """
        Normalizes the properties dictionary to ensure consistent structure.
        :param properties: The properties to be normalized.
        """
        type_value = properties.pop("type", None)

        if isinstance(type_value, str):
            properties["type"] = type_value
            return properties

        if isinstance(type_value, list) and len(type_value) == 0:
            raise InvalidSchemaException(
                "Invalid schema: 'type' list cannot be empty",
                invalid_field=str(properties),
            )

        if isinstance(type_value, list) and len(type_value) == 1:
            properties["type"] = type_value[0]
            return properties

        if isinstance(type_value, list):
            properties["anyOf"] = [{"type": t} for t in type_value]
            return properties

        return properties

    @classmethod
    def _get_impl(cls, properties: JSONSchema) -> type[Self]:
        for subcls in cls.__subclasses__():
            schema_type, schema_value = subcls._get_schema_type()

            if schema_type not in properties:
                continue

            if schema_value is None or schema_value == properties[schema_type]:  # type: ignore
                return subcls

        raise InvalidSchemaException(
            "No suitable type parser found", invalid_field=str(properties)
        )

    @classmethod
    def _get_schema_type(cls) -> tuple[str, str | None]:
        if cls.json_schema_type is None:
            raise RuntimeError(
                f"TypeParser: json_schema_type not defined for subclass {cls.__name__}"
            )

        schema_definition = cls.json_schema_type.split(":")

        if len(schema_definition) == 1:
            return schema_definition[0], None

        return schema_definition[0], schema_definition[1]

    def mappings_properties_builder(
        self, properties, **kwargs: Unpack[TypeParserOptions]
    ) -> dict[str, Any]:
        if not kwargs.get("required", False):
            properties["default"] = properties.get("default", None)

        mappings = self.default_mappings | self.type_mappings

        return {
            mappings[key]: value for key, value in properties.items() if key in mappings
        }

    @staticmethod
    def _validate_default(field_type: T, field_prop: dict) -> bool:
        value = field_prop.get("default")

        if value is None and field_prop.get("default_factory") is not None:
            value = field_prop["default_factory"]()

        if value is None:
            return True

        return GenericTypeParser._is_valid_value(field_type, field_prop, value)

    @staticmethod
    def _validate_examples(field_type: T, field_prop: dict) -> bool:
        examples = field_prop.get("examples")

        if examples is None:
            return True

        if not isinstance(examples, list):
            return False

        return all(
            GenericTypeParser._is_valid_value(field_type, field_prop, e)
            for e in examples
        )

    @staticmethod
    def _is_valid_value(field_type: T, field_prop: dict, value: Any) -> bool:
        try:
            field = Annotated[field_type, Field(**field_prop)]  # type: ignore
            TypeAdapter(field).validate_python(value)
        except Exception as _:
            return False

        return True



from ._type_parser import GenericTypeParser
from .allof_type_parser import AllOfTypeParser
from .anyof_type_parser import AnyOfTypeParser
from .array_type_parser import ArrayTypeParser
from .boolean_type_parser import BooleanTypeParser
from .const_type_parser import ConstTypeParser
from .enum_type_parser import EnumTypeParser
from .float_type_parser import FloatTypeParser
from .int_type_parser import IntTypeParser
from .null_type_parser import NullTypeParser
from .object_type_parser import ObjectTypeParser
from .oneof_type_parser import OneOfTypeParser
from .ref_type_parser import RefTypeParser
from .string_type_parser import StringTypeParser


__all__ = [
    "GenericTypeParser",
    "EnumTypeParser",
    "ConstTypeParser",
    "AllOfTypeParser",
    "AnyOfTypeParser",
    "ArrayTypeParser",
    "BooleanTypeParser",
    "FloatTypeParser",
    "IntTypeParser",
    "NullTypeParser",
    "ObjectTypeParser",
    "OneOfTypeParser",
    "StringTypeParser",
    "RefTypeParser",
]



from .internal_assertion_exception import InternalAssertionException
from .invalid_schema_exception import InvalidSchemaException
from .unsupported_schema_exception import UnsupportedSchemaException


__all__ = [
    "InternalAssertionException",
    "InvalidSchemaException",
    "UnsupportedSchemaException",
]



from .json_schema_type import (
    JSONSchema,
    JSONSchemaNativeTypes,
    JSONSchemaType,
    JSONType,
)
from .type_parser_options import RefCacheDict, TypeParserOptions


__all__ = [
    "JSONSchemaType",
    "JSONSchemaNativeTypes",
    "JSONType",
    "JSONSchema",
    "RefCacheDict",
    "TypeParserOptions",
]



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.json_schema_type import JSONSchema
from jambo.types.type_parser_options import TypeParserOptions

from typing_extensions import Unpack


class AllOfTypeParser(GenericTypeParser):
    mapped_type = any

    json_schema_type = "allOf"

    def from_properties_impl(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ):
        sub_properties = properties.get("allOf", [])

        root_type = properties.get("type")
        if root_type is not None:
            for sub_property in sub_properties:
                sub_property["type"] = root_type

        parser = self._get_type_parser(sub_properties)

        combined_properties = self._rebuild_properties_from_subproperties(
            sub_properties
        )

        if (examples := properties.get("examples")) is not None:
            combined_properties["examples"] = examples

        return parser().from_properties_impl(name, combined_properties, **kwargs)

    @staticmethod
    def _get_type_parser(
        sub_properties: list[JSONSchema],
    ) -> type[GenericTypeParser]:
        if not sub_properties:
            raise InvalidSchemaException(
                "'allOf' must contain at least one schema", invalid_field="allOf"
            )

        parsers: set[type[GenericTypeParser]] = set(
            GenericTypeParser._get_impl(sub_property) for sub_property in sub_properties
        )
        if len(parsers) != 1:
            raise InvalidSchemaException(
                "All sub-schemas in 'allOf' must resolve to the same type",
                invalid_field="allOf",
            )

        return parsers.pop()

    @staticmethod
    def _rebuild_properties_from_subproperties(
        sub_properties: list[JSONSchema],
    ) -> JSONSchema:
        properties: JSONSchema = {}
        for subProperty in sub_properties:
            for name, prop in subProperty.items():
                if name not in properties:
                    properties[name] = prop  # type: ignore
                else:
                    # Merge properties if they exist in both sub-properties
                    properties[name] = AllOfTypeParser._validate_prop(  # type: ignore
                        name,
                        properties[name],  # type: ignore
                        prop,
                    )
        return properties

    @staticmethod
    def _validate_prop(prop_name, old_value, new_value):
        if prop_name == "description":
            return f"{old_value} | {new_value}"

        if prop_name == "default":
            if old_value != new_value:
                raise InvalidSchemaException(
                    f"Conflicting defaults for '{prop_name}'", invalid_field=prop_name
                )
            return old_value

        if prop_name == "required":
            return old_value + new_value

        if prop_name in ("maxLength", "maximum", "exclusiveMaximum"):
            return old_value if old_value > new_value else new_value

        if prop_name in ("minLength", "minimum", "exclusiveMinimum"):
            return old_value if old_value < new_value else new_value

        if prop_name == "properties":
            for key, value in new_value.items():
                if key not in old_value:
                    old_value[key] = value
                    continue

                for sub_key, sub_value in value.items():
                    if sub_key not in old_value[key]:
                        old_value[key][sub_key] = sub_value
                    else:
                        # Merge properties if they exist in both sub-properties
                        old_value[key][sub_key] = AllOfTypeParser._validate_prop(
                            sub_key, old_value[key][sub_key], sub_value
                        )

        # Handle other properties by just returning the first valued
        return old_value



from jambo.exceptions import InvalidSchemaException
from jambo.parser import FloatTypeParser

from unittest import TestCase


class TestFloatTypeParser(TestCase):
    def test_float_parser_no_options(self):
        parser = FloatTypeParser()

        properties = {"type": "number"}

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, float)
        self.assertEqual(type_validator, {"default": None})

    def test_float_parser_with_options(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "maximum": 10.5,
            "minimum": 1.0,
            "multipleOf": 0.5,
            "examples": [1.5, 2.5],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, float)
        self.assertEqual(type_validator["le"], 10.5)
        self.assertEqual(type_validator["ge"], 1.0)
        self.assertEqual(type_validator["multiple_of"], 0.5)
        self.assertEqual(type_validator["examples"], [1.5, 2.5])

    def test_float_parser_with_default(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "default": 5.0,
            "maximum": 10.5,
            "minimum": 1.0,
            "multipleOf": 0.5,
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, float)
        self.assertEqual(type_validator["default"], 5.0)
        self.assertEqual(type_validator["le"], 10.5)
        self.assertEqual(type_validator["ge"], 1.0)
        self.assertEqual(type_validator["multiple_of"], 0.5)

    def test_float_parser_with_default_invalid_type(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "default": "invalid",  # Invalid default value
            "maximum": 10.5,
            "minimum": 1.0,
            "multipleOf": 0.5,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_float_parser_with_default_invalid_maximum(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "default": 15.0,
            "maximum": 10.5,
            "minimum": 1.0,
            "multipleOf": 0.5,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_float_parser_with_default_invalid_minimum(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "default": -5.0,
            "maximum": 10.5,
            "minimum": 1.0,
            "multipleOf": 0.5,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_float_parser_with_default_invalid_exclusive_maximum(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "default": 10.5,
            "exclusiveMaximum": 10.5,
            "minimum": 1.0,
            "multipleOf": 0.5,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_float_parser_with_default_invalid_exclusive_minimum(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "default": 1.0,
            "maximum": 10.5,
            "exclusiveMinimum": 1.0,
            "multipleOf": 0.5,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_float_parser_with_default_invalid_multiple(self):
        parser = FloatTypeParser()

        properties = {
            "type": "number",
            "default": 5.0,
            "maximum": 10.5,
            "minimum": 1.0,
            "multipleOf": 2.0,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)



from jambo.exceptions import InvalidSchemaException
from jambo.parser import StringTypeParser

from pydantic import AnyUrl, EmailStr

import unittest
from datetime import date, datetime, time, timedelta, timezone
from ipaddress import IPv4Address, IPv6Address, ip_address
from unittest import TestCase
from uuid import UUID


class TestStringTypeParser(TestCase):
    def test_string_parser_no_options(self):
        parser = StringTypeParser()

        properties = {"type": "string"}

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, str)

    def test_string_parser_with_options(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "maxLength": 10,
            "minLength": 1,
            "pattern": "^[a-zA-Z]+$",
            "examples": ["test", "TEST"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, str)
        self.assertEqual(type_validator["max_length"], 10)
        self.assertEqual(type_validator["min_length"], 1)
        self.assertEqual(type_validator["pattern"], "^[a-zA-Z]+$")
        self.assertEqual(type_validator["examples"], ["test", "TEST"])

    def test_string_parser_with_default_value(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "default": "default_value",
            "maxLength": 20,
            "minLength": 5,
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, str)
        self.assertEqual(type_validator["default"], "default_value")
        self.assertEqual(type_validator["max_length"], 20)
        self.assertEqual(type_validator["min_length"], 5)

    def test_string_parser_with_invalid_default_value_type(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "default": 12345,  # Invalid default value
            "maxLength": 20,
            "minLength": 5,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_string_parser_with_default_invalid_maxlength(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "default": "default_value",
            "maxLength": 2,
            "minLength": 1,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_string_parser_with_default_invalid_minlength(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "default": "a",
            "maxLength": 20,
            "minLength": 2,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_string_parser_with_email_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "email",
            "examples": ["test@example.com"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, EmailStr)
        self.assertEqual(type_validator["examples"], ["test@example.com"])

    def test_string_parser_with_uri_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "uri",
            "examples": ["test://domain/resource"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, AnyUrl)
        self.assertEqual(type_validator["examples"], ["test://domain/resource"])

    def test_string_parser_with_ip_formats(self):
        parser = StringTypeParser()

        formats = {"ipv4": IPv4Address, "ipv6": IPv6Address}
        examples = {"ipv4": ["192.168.1.1"], "ipv6": ["::1"]}

        for ip_format, expected_type in formats.items():
            example = examples[ip_format]

            properties = {
                "type": "string",
                "format": ip_format,
                "examples": example,
            }

            type_parsing, type_validator = parser.from_properties(
                "placeholder", properties
            )

            self.assertEqual(type_parsing, expected_type)
            self.assertEqual(
                type_validator["examples"], [ip_address(e) for e in example]
            )

    def test_string_parser_with_uuid_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "uuid",
            "examples": ["ab71aaf4-ab6e-43cd-a369-cebdd9f7a4c6"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, UUID)
        self.assertEqual(
            type_validator["examples"], [UUID("ab71aaf4-ab6e-43cd-a369-cebdd9f7a4c6")]
        )

    def test_string_parser_with_time_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "time",
            "examples": ["14:30:00", "09:15:30.500", "10:00:00+02:00"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, time)
        self.assertEqual(
            type_validator["examples"],
            [
                time(hour=14, minute=30, second=0),
                time(hour=9, minute=15, second=30, microsecond=500_000),
                time(hour=10, minute=0, second=0, tzinfo=timezone(timedelta(hours=2))),
            ],
        )

    def test_string_parser_with_pattern_based_formats(self):
        parser = StringTypeParser()

        format_types = {
            "hostname": "example.com",
        }

        for format_type, example_type in format_types.items():
            properties = {
                "type": "string",
                "format": format_type,
                "examples": [example_type],
            }

            type_parsing, type_validator = parser.from_properties(
                "placeholder", properties
            )

            self.assertEqual(type_parsing, str)
            self.assertIn("pattern", type_validator)
            self.assertEqual(
                type_validator["pattern"], parser.format_pattern_mapping[format_type]
            )
            self.assertEqual(type_validator["examples"], [example_type])

    def test_string_parser_with_unsupported_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "unsupported-format",
        }

        with self.assertRaises(InvalidSchemaException) as context:
            parser.from_properties("placeholder", properties)

        self.assertEqual(
            str(context.exception),
            "Invalid JSON Schema: Unsupported string format: unsupported-format (invalid field: format)",
        )

    def test_string_parser_with_date_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "date",
            "examples": ["2025-11-17", "1999-12-31", "2000-01-01"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, date)
        self.assertEqual(
            type_validator["examples"],
            [
                date(year=2025, month=11, day=17),
                date(year=1999, month=12, day=31),
                date(year=2000, month=1, day=1),
            ],
        )

    def test_string_parser_with_datetime_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "date-time",
            "examples": [
                "2025-11-17T11:15:00",
                "2025-11-17T11:15:00+01:00",
                "2025-11-17T11:15:00.123456-05:00",
            ],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, datetime)
        self.assertEqual(
            type_validator["examples"],
            [
                datetime(year=2025, month=11, day=17, hour=11, minute=15, second=0),
                datetime(
                    year=2025,
                    month=11,
                    day=17,
                    hour=11,
                    minute=15,
                    second=0,
                    tzinfo=timezone(timedelta(hours=1)),
                ),
                datetime(
                    year=2025,
                    month=11,
                    day=17,
                    hour=11,
                    minute=15,
                    second=0,
                    microsecond=123456,
                    tzinfo=timezone(timedelta(hours=-5)),
                ),
            ],
        )

    def test_string_parser_with_invalid_example_value(self):
        with self.assertRaises(InvalidSchemaException):
            StringTypeParser().from_properties(
                "placeholder",
                {
                    "type": "string",
                    "format": "email",
                    "examples": ["invalid-email"],
                },
            )

    @unittest.skip("Duration parsing not yet implemented")
    def test_string_parser_with_timedelta_format(self):
        parser = StringTypeParser()

        properties = {
            "type": "string",
            "format": "duration",
            "examples": ["P1Y2M3DT4H5M6S", "PT30M", "P7D", "PT0.5S"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, timedelta)
        self.assertEqual(
            type_validator["examples"],
            [
                timedelta(days=7),
                timedelta(minutes=30),
                timedelta(hours=4, minutes=5, seconds=6),
                timedelta(seconds=0.5),
            ],
        )



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.type_parser_options import TypeParserOptions

from typing_extensions import Unpack


class BooleanTypeParser(GenericTypeParser):
    mapped_type = bool

    json_schema_type = "type:boolean"

    type_mappings = {
        "default": "default",
    }

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        mapped_properties = self.mappings_properties_builder(properties, **kwargs)

        default_value = properties.get("default")
        if default_value is not None and not isinstance(default_value, bool):
            raise InvalidSchemaException(
                f"Default value for {name} must be a boolean.",
                invalid_field="default",
            )

        return bool, mapped_properties



from jambo.exceptions import InvalidSchemaException
from jambo.parser import ConstTypeParser

from typing_extensions import Annotated, Literal, get_args, get_origin

from unittest import TestCase


class TestConstTypeParser(TestCase):
    def test_const_type_parser_hashable_value(self):
        """Test const parser with hashable values (uses Literal)"""
        parser = ConstTypeParser()

        expected_const_value = "United States of America"
        properties = {"const": expected_const_value, "examples": [expected_const_value]}

        parsed_type, parsed_properties = parser.from_properties_impl(
            "country", properties
        )

        # Check that we get a Literal type for hashable values
        self.assertEqual(get_origin(parsed_type), Literal)
        self.assertEqual(get_args(parsed_type), (expected_const_value,))

        self.assertEqual(parsed_properties["default"], expected_const_value)
        self.assertEqual(parsed_properties["examples"], [expected_const_value])

    def test_const_type_parser_non_hashable_value(self):
        """Test const parser with non-hashable values (uses Annotated with validator)"""
        parser = ConstTypeParser()

        expected_const_value = [1, 2, 3]  # Lists are not hashable
        properties = {"const": expected_const_value, "examples": [expected_const_value]}

        parsed_type, parsed_properties = parser.from_properties_impl(
            "list_const", properties
        )

        # Check that we get an Annotated type for non-hashable values
        self.assertEqual(get_origin(parsed_type), Annotated)
        self.assertIn(list, get_args(parsed_type))

        self.assertEqual(parsed_properties["default"], expected_const_value)
        self.assertEqual(parsed_properties["examples"], [expected_const_value])

    def test_const_type_parser_integer_value(self):
        """Test const parser with integer values (uses Literal)"""
        parser = ConstTypeParser()

        expected_const_value = 42
        properties = {"const": expected_const_value, "examples": [expected_const_value]}

        parsed_type, parsed_properties = parser.from_properties_impl(
            "int_const", properties
        )

        # Check that we get a Literal type for hashable values
        self.assertEqual(get_origin(parsed_type), Literal)
        self.assertEqual(get_args(parsed_type), (expected_const_value,))

        self.assertEqual(parsed_properties["default"], expected_const_value)
        self.assertEqual(parsed_properties["examples"], [expected_const_value])

    def test_const_type_parser_boolean_value(self):
        """Test const parser with boolean values (uses Literal)"""
        parser = ConstTypeParser()

        expected_const_value = True
        properties = {"const": expected_const_value, "examples": [expected_const_value]}

        parsed_type, parsed_properties = parser.from_properties_impl(
            "bool_const", properties
        )

        # Check that we get a Literal type for hashable values
        self.assertEqual(get_origin(parsed_type), Literal)
        self.assertEqual(get_args(parsed_type), (expected_const_value,))

        self.assertEqual(parsed_properties["default"], expected_const_value)
        self.assertEqual(parsed_properties["examples"], [expected_const_value])

    def test_const_type_parser_invalid_properties(self):
        parser = ConstTypeParser()

        expected_const_value = "United States of America"
        properties = {"notConst": expected_const_value}

        with self.assertRaises(InvalidSchemaException) as context:
            parser.from_properties_impl("invalid_country", properties)

        self.assertIn(
            "Const type invalid_country must have 'const' property defined",
            str(context.exception),
        )

    def test_const_type_parser_invalid_const_value(self):
        parser = ConstTypeParser()

        properties = {"const": {}}

        with self.assertRaises(InvalidSchemaException) as context:
            parser.from_properties_impl("invalid_country", properties)

        self.assertIn(
            "Const type invalid_country must have 'const' value of allowed types",
            str(context.exception),
        )



from jambo import SchemaConverter
from jambo.exceptions import InvalidSchemaException
from jambo.parser.oneof_type_parser import OneOfTypeParser

from pydantic import ValidationError

from unittest import TestCase


class TestOneOfTypeParser(TestCase):
    def test_oneof_raises_on_invalid_property(self):
        with self.assertRaises(InvalidSchemaException):
            OneOfTypeParser().from_properties_impl(
                "test_field",
                {
                    # Invalid schema, should have property "oneOf"
                },
                required=True,
                context={},
                ref_cache={},
            )

        with self.assertRaises(InvalidSchemaException):
            OneOfTypeParser().from_properties_impl(
                "test_field",
                {
                    "oneOf": [],  # should throw because oneOf must be a list with at least one item
                },
                required=True,
                context={},
                ref_cache={},
            )

        with self.assertRaises(InvalidSchemaException):
            SchemaConverter.build(
                {
                    "title": "Test",
                    "type": "object",
                    "properties": {
                        "value": {
                            "oneOf": [],  # should throw because oneOf requires at least one schema
                        }
                    },
                }
            )

    def test_oneof_basic_integer_and_string(self):
        schema = {
            "title": "Person",
            "description": "A person with an ID that can be either an integer or a formatted string",
            "type": "object",
            "properties": {
                "id": {
                    "oneOf": [
                        {"type": "integer", "minimum": 1},
                        {"type": "string", "pattern": "^[A-Z]{2}[0-9]{4}$"},
                    ]
                },
            },
            "required": ["id"],
        }

        Model = SchemaConverter.build(schema)

        obj1 = Model(id=123)
        self.assertEqual(obj1.id, 123)

        obj2 = Model(id="AB1234")
        self.assertEqual(obj2.id, "AB1234")

    def test_oneof_validation_failures(self):
        schema = {
            "title": "Person",
            "type": "object",
            "properties": {
                "id": {
                    "oneOf": [
                        {"type": "integer", "minimum": 1},
                        {"type": "string", "pattern": "^[A-Z]{2}[0-9]{4}$"},
                    ]
                },
            },
            "required": ["id"],
        }

        Model = SchemaConverter.build(schema)

        with self.assertRaises(ValidationError):
            Model(id=-5)

        with self.assertRaises(ValidationError):
            Model(id="invalid")

        with self.assertRaises(ValidationError):
            Model(id=123.45)

    def test_oneof_with_conflicting_schemas(self):
        schema = {
            "title": "Value",
            "type": "object",
            "properties": {
                "data": {
                    "oneOf": [
                        {"type": "number", "multipleOf": 2},
                        {"type": "number", "multipleOf": 3},
                    ]
                },
            },
            "required": ["data"],
        }

        Model = SchemaConverter.build(schema)

        obj1 = Model(data=4)
        self.assertEqual(obj1.data, 4)

        obj2 = Model(data=9)
        self.assertEqual(obj2.data, 9)

        with self.assertRaises(ValidationError) as cm:
            Model(data=6)
        self.assertIn("matches multiple oneOf schemas", str(cm.exception))

        with self.assertRaises(ValidationError):
            Model(data=5)

    def test_oneof_with_objects(self):
        schema = {
            "title": "Contact",
            "type": "object",
            "properties": {
                "contact_info": {
                    "oneOf": [
                        {
                            "type": "object",
                            "properties": {
                                "email": {"type": "string", "format": "email"}
                            },
                            "required": ["email"],
                            "additionalProperties": False,
                        },
                        {
                            "type": "object",
                            "properties": {
                                "phone": {"type": "string", "pattern": "^[0-9-]+$"}
                            },
                            "required": ["phone"],
                            "additionalProperties": False,
                        },
                    ]
                },
            },
            "required": ["contact_info"],
        }

        Model = SchemaConverter.build(schema)

        obj1 = Model(contact_info={"email": "user@example.com"})
        self.assertEqual(obj1.contact_info.email, "user@example.com")

        obj2 = Model(contact_info={"phone": "123-456-7890"})
        self.assertEqual(obj2.contact_info.phone, "123-456-7890")

        with self.assertRaises(ValidationError):
            Model(contact_info={"email": "user@example.com", "phone": "123-456-7890"})

    def test_oneof_with_discriminator_basic(self):
        schema = {
            "title": "Pet",
            "type": "object",
            "properties": {
                "pet": {
                    "oneOf": [
                        {
                            "type": "object",
                            "properties": {
                                "type": {"const": "cat"},
                                "meows": {"type": "boolean"},
                            },
                            "required": ["type", "meows"],
                        },
                        {
                            "type": "object",
                            "properties": {
                                "type": {"const": "dog"},
                                "barks": {"type": "boolean"},
                            },
                            "required": ["type", "barks"],
                        },
                    ],
                    "discriminator": {"propertyName": "type"},
                }
            },
            "required": ["pet"],
        }

        Model = SchemaConverter.build(schema)

        cat = Model(pet={"type": "cat", "meows": True})
        self.assertEqual(cat.pet.type, "cat")
        self.assertEqual(cat.pet.meows, True)

        dog = Model(pet={"type": "dog", "barks": False})
        self.assertEqual(dog.pet.type, "dog")
        self.assertEqual(dog.pet.barks, False)

        with self.assertRaises(ValidationError):
            Model(pet={"type": "cat", "barks": True})

        with self.assertRaises(ValidationError):
            Model(pet={"type": "bird", "flies": True})

    def test_oneof_with_invalid_types(self):
        with self.assertRaises(InvalidSchemaException):
            SchemaConverter.build(
                {
                    "title": "Pet",
                    "type": "object",
                    "properties": {
                        "pet": {
                            "oneOf": [
                                {
                                    "type": "number",
                                },
                                {
                                    "type": "string",
                                },
                            ],
                            "discriminator": {"propertyName": "type"},
                        }
                    },
                    "required": ["pet"],
                }
            )

    def test_oneof_with_discriminator_mapping(self):
        schema = {
            "title": "Vehicle",
            "type": "object",
            "properties": {
                "vehicle": {
                    "oneOf": [
                        {
                            "type": "object",
                            "properties": {
                                "vehicle_type": {"const": "car"},
                                "doors": {
                                    "type": "integer",
                                    "minimum": 2,
                                    "maximum": 4,
                                },
                            },
                            "required": ["vehicle_type", "doors"],
                        },
                        {
                            "type": "object",
                            "properties": {
                                "vehicle_type": {"const": "motorcycle"},
                                "engine_size": {"type": "number", "minimum": 125},
                            },
                            "required": ["vehicle_type", "engine_size"],
                        },
                    ],
                    "discriminator": {
                        "propertyName": "vehicle_type",
                        "mapping": {
                            "car": "#/properties/vehicle/oneOf/0",
                            "motorcycle": "#/properties/vehicle/oneOf/1",
                        },
                    },
                }
            },
            "required": ["vehicle"],
        }

        Model = SchemaConverter.build(schema)

        car = Model(vehicle={"vehicle_type": "car", "doors": 4})
        self.assertEqual(car.vehicle.vehicle_type, "car")
        self.assertEqual(car.vehicle.doors, 4)

        motorcycle = Model(vehicle={"vehicle_type": "motorcycle", "engine_size": 600.0})
        self.assertEqual(motorcycle.vehicle.vehicle_type, "motorcycle")
        self.assertEqual(motorcycle.vehicle.engine_size, 600.0)

    def test_oneof_with_discriminator_invalid_values(self):
        schema = {
            "title": "Shape",
            "type": "object",
            "properties": {
                "shape": {
                    "oneOf": [
                        {
                            "type": "object",
                            "properties": {
                                "type": {"const": "circle"},
                                "radius": {"type": "number", "minimum": 0},
                            },
                            "required": ["type", "radius"],
                        },
                        {
                            "type": "object",
                            "properties": {
                                "type": {"const": "square"},
                                "side": {"type": "number", "minimum": 0},
                            },
                            "required": ["type", "side"],
                        },
                    ],
                    "discriminator": {"propertyName": "type"},
                }
            },
            "required": ["shape"],
        }

        Model = SchemaConverter.build(schema)

        with self.assertRaises(ValidationError):
            Model(shape={"type": "triangle", "base": 5, "height": 3})

        with self.assertRaises(ValidationError):
            Model(shape={"type": "circle", "side": 5})

        with self.assertRaises(ValidationError):
            Model(shape={"radius": 5})

    def test_oneof_missing_properties(self):
        schema = {
            "title": "Test",
            "type": "object",
            "properties": {
                "value": {
                    "notOneOf": [
                        {"type": "string"},
                        {"type": "integer"},
                    ]
                },
            },
        }

        with self.assertRaises(InvalidSchemaException):
            SchemaConverter.build(schema)

    def test_oneof_invalid_properties(self):
        schema = {
            "title": "Test",
            "type": "object",
            "properties": {
                "value": {"oneOf": None},
            },
        }

        with self.assertRaises(InvalidSchemaException):
            SchemaConverter.build(schema)

    def test_oneof_with_default_value(self):
        schema = {
            "title": "Test",
            "type": "object",
            "properties": {
                "value": {
                    "oneOf": [
                        {"type": "string"},
                        {"type": "integer"},
                    ],
                    "default": "test",
                },
            },
        }

        Model = SchemaConverter.build(schema)
        obj = Model()
        self.assertEqual(obj.value, "test")

    def test_oneof_with_invalid_default_value(self):
        schema = {
            "title": "Test",
            "type": "object",
            "properties": {
                "value": {
                    "oneOf": [
                        {"type": "string", "minLength": 5},
                        {"type": "integer", "minimum": 10},
                    ],
                    "default": "hi",
                },
            },
        }

        with self.assertRaises(InvalidSchemaException):
            SchemaConverter.build(schema)

    def test_oneof_discriminator_without_property_name(self):
        # Should throw because the spec determines propertyName is required for discriminator
        with self.assertRaises(InvalidSchemaException):
            SchemaConverter.build(
                {
                    "title": "Test",
                    "type": "object",
                    "properties": {
                        "value": {
                            "oneOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "type": {"const": "a"},
                                        "value": {"type": "string"},
                                    },
                                },
                                {
                                    "type": "object",
                                    "properties": {
                                        "type": {"const": "b"},
                                        "value": {"type": "integer"},
                                    },
                                },
                            ],
                            "discriminator": {},  # discriminator without propertyName
                        }
                    },
                }
            )

    def test_oneof_discriminator_with_invalid_discriminator(self):
        # Should throw because a valid discriminator is required
        with self.assertRaises(InvalidSchemaException):
            SchemaConverter.build(
                {
                    "title": "Test",
                    "type": "object",
                    "properties": {
                        "value": {
                            "oneOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "type": {"const": "a"},
                                        "value": {"type": "string"},
                                    },
                                },
                                {
                                    "type": "object",
                                    "properties": {
                                        "type": {"const": "b"},
                                        "value": {"type": "integer"},
                                    },
                                },
                            ],
                            "discriminator": "invalid",  # discriminator without propertyName
                        }
                    },
                }
            )

    def test_oneof_overlapping_strings_from_docs(self):
        """Test the overlapping strings example from documentation"""
        schema = {
            "title": "SimpleExample",
            "type": "object",
            "properties": {
                "value": {
                    "oneOf": [
                        {"type": "string", "maxLength": 6},
                        {"type": "string", "minLength": 4},
                    ]
                }
            },
            "required": ["value"],
        }

        Model = SchemaConverter.build(schema)

        # Valid: Short string (matches first schema only)
        obj1 = Model(value="hi")
        self.assertEqual(obj1.value, "hi")

        # Valid: Long string (matches second schema only)
        obj2 = Model(value="very long string")
        self.assertEqual(obj2.value, "very long string")

        # Invalid: Medium string (matches BOTH schemas - violates oneOf)
        with self.assertRaises(ValidationError) as cm:
            Model(value="hello")  # 5 chars: matches maxLength=6 AND minLength=4

        self.assertIn("matches multiple oneOf schemas", str(cm.exception))

    def test_oneof_shapes_discriminator_from_docs(self):
        """Test the shapes discriminator example from documentation"""
        schema = {
            "title": "Shape",
            "type": "object",
            "properties": {
                "shape": {
                    "oneOf": [
                        {
                            "type": "object",
                            "properties": {
                                "type": {"const": "circle"},
                                "radius": {"type": "number", "minimum": 0},
                            },
                            "required": ["type", "radius"],
                        },
                        {
                            "type": "object",
                            "properties": {
                                "type": {"const": "rectangle"},
                                "width": {"type": "number", "minimum": 0},
                                "height": {"type": "number", "minimum": 0},
                            },
                            "required": ["type", "width", "height"],
                        },
                    ],
                    "discriminator": {"propertyName": "type"},
                }
            },
            "required": ["shape"],
        }

        Model = SchemaConverter.build(schema)

        # Valid: Circle
        circle = Model(shape={"type": "circle", "radius": 5.0})
        self.assertEqual(circle.shape.type, "circle")
        self.assertEqual(circle.shape.radius, 5.0)

        # Valid: Rectangle
        rectangle = Model(shape={"type": "rectangle", "width": 10, "height": 20})
        self.assertEqual(rectangle.shape.type, "rectangle")
        self.assertEqual(rectangle.shape.width, 10)
        self.assertEqual(rectangle.shape.height, 20)

        # Invalid: Wrong properties for the type
        with self.assertRaises(ValidationError):
            Model(shape={"type": "circle", "width": 10})

    def test_oneof_with_examples(self):
        schema = {
            "title": "ExampleTest",
            "type": "object",
            "properties": {
                "value": {
                    "oneOf": [
                        {
                            "type": "string",
                            "examples": ["example1", "example2"],
                        },
                        {
                            "type": "integer",
                            "examples": [1, 2, 3],
                        },
                    ]
                }
            },
            "required": ["value"],
        }

        Model = SchemaConverter.build(schema)

        # Since Pydantic does not natively support oneOf and the validation
        # is done via a custom validator, the `value` is represented using `anyOf`
        model_schema = Model.model_json_schema()

        self.assertEqual(
            model_schema["properties"]["value"]["anyOf"][0]["examples"],
            ["example1", "example2"],
        )

        self.assertEqual(
            model_schema["properties"]["value"]["anyOf"][1]["examples"],
            [1, 2, 3],
        )

    def test_oneof_with_root_examples(self):
        schema = {
            "title": "ExampleTest",
            "type": "object",
            "properties": {
                "value": {
                    "oneOf": [
                        {
                            "type": "string",
                        },
                        {
                            "type": "integer",
                        },
                    ],
                    "examples": ["example1", 2],
                }
            },
            "required": ["value"],
        }

        Model = SchemaConverter.build(schema)

        # Since Pydantic does not natively support oneOf and the validation
        # is done via a custom validator, the `value` is represented using `anyOf`
        model_schema = Model.model_json_schema()

        self.assertEqual(
            model_schema["properties"]["value"]["examples"],
            ["example1", 2],
        )



from jambo.exceptions import InvalidSchemaException
from jambo.parser import EnumTypeParser

from enum import Enum
from unittest import TestCase


class TestEnumTypeParser(TestCase):
    def test_enum_type_parser_throws_enum_not_defined(self):
        parser = EnumTypeParser()

        schema = {}

        with self.assertRaises(InvalidSchemaException):
            parsed_type, parsed_properties = parser.from_properties_impl(
                "TestEnum",
                schema,
            )

    def test_enum_type_parser_throws_enum_not_list(self):
        parser = EnumTypeParser()

        schema = {
            "enum": "not_a_list",
        }

        with self.assertRaises(InvalidSchemaException):
            parsed_type, parsed_properties = parser.from_properties_impl(
                "TestEnum",
                schema,
            )

    def test_enum_type_parser_creates_enum(self):
        parser = EnumTypeParser()

        schema = {
            "enum": ["value1", "value2", "value3"],
        }

        parsed_type, parsed_properties = parser.from_properties_impl(
            "TestEnum",
            schema,
        )

        self.assertIsInstance(parsed_type, type)
        self.assertTrue(issubclass(parsed_type, Enum))
        self.assertEqual(
            set(parsed_type.__members__.keys()), {"VALUE1", "VALUE2", "VALUE3"}
        )
        self.assertEqual(parsed_properties, {"default": None})

    def test_enum_type_parser_creates_enum_with_default(self):
        parser = EnumTypeParser()

        schema = {
            "enum": ["value1", "value2", "value3"],
            "default": "value2",
        }

        parsed_type, parsed_properties = parser.from_properties_impl(
            "TestEnum",
            schema,
        )

        self.assertIsInstance(parsed_type, type)
        self.assertTrue(issubclass(parsed_type, Enum))
        self.assertEqual(
            set(parsed_type.__members__.keys()), {"VALUE1", "VALUE2", "VALUE3"}
        )
        self.assertEqual(parsed_properties["default"].value, "value2")

    def test_enum_type_parser_throws_invalid_default(self):
        parser = EnumTypeParser()

        schema = {
            "enum": ["value1", "value2", "value3"],
            "default": "invalid_value",
        }

        with self.assertRaises(ValueError):
            parser.from_properties_impl("TestEnum", schema)

    def test_enum_type_parser_throws_invalid_enum_value(self):
        parser = EnumTypeParser()

        schema = {
            "enum": ["value1", 42, dict()],
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties_impl("TestEnum", schema)

    def test_enum_type_parser_creates_enum_with_examples(self):
        parser = EnumTypeParser()

        schema = {
            "enum": ["value1", "value2", "value3"],
            "examples": ["value1", "value3"],
        }

        parsed_type, parsed_properties = parser.from_properties_impl(
            "TestEnum",
            schema,
        )

        self.assertIsInstance(parsed_type, type)
        self.assertTrue(issubclass(parsed_type, Enum))
        self.assertEqual(
            set(parsed_type.__members__.keys()), {"VALUE1", "VALUE2", "VALUE3"}
        )
        self.assertEqual(parsed_properties["default"], None)
        self.assertEqual(
            parsed_properties["examples"],
            [getattr(parsed_type, "VALUE1"), getattr(parsed_type, "VALUE3")],
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "jambo"
copyright = "2025, Vitor Hideyoshi"
author = "Vitor Hideyoshi"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",  # <-- needed
]


templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]


# -- Options for autodoc -----------------------------------------------------
add_module_names = False
python_use_unqualified_type_names = True


autodoc_typehints = "both"



from jambo.exceptions.invalid_schema_exception import InvalidSchemaException

from unittest import TestCase


class TestInternalAssertionException(TestCase):
    def test_inheritance(self):
        self.assertTrue(issubclass(InvalidSchemaException, ValueError))

    def test_message(self):
        message = "This is an internal assertion error."

        expected_message = f"Invalid JSON Schema: {message}"

        with self.assertRaises(InvalidSchemaException) as ctx:
            raise InvalidSchemaException(message)

        self.assertEqual(str(ctx.exception), expected_message)

    def test_invalid_field(self):
        message = "This is an internal assertion error."
        invalid_field = "testField"

        expected_message = (
            f"Invalid JSON Schema: {message} (invalid field: {invalid_field})"
        )

        with self.assertRaises(InvalidSchemaException) as ctx:
            raise InvalidSchemaException(message, invalid_field=invalid_field)

        self.assertEqual(str(ctx.exception), expected_message)

    def test_cause(self):
        message = "This is an internal assertion error."
        cause = ValueError("Underlying cause")

        expected_message = (
            f"Invalid JSON Schema: {message} (caused by ValueError: Underlying cause)"
        )

        with self.assertRaises(InvalidSchemaException) as ctx:
            raise InvalidSchemaException(message, cause=cause)

        self.assertEqual(str(ctx.exception), expected_message)



from jambo.exceptions.unsupported_schema_exception import UnsupportedSchemaException

from unittest import TestCase


class TestUnsupportedSchemaException(TestCase):
    def test_inheritance(self):
        self.assertTrue(issubclass(UnsupportedSchemaException, ValueError))

    def test_message(self):
        message = "This is an internal assertion error."

        expected_message = f"Unsupported JSON Schema: {message}"

        with self.assertRaises(UnsupportedSchemaException) as ctx:
            raise UnsupportedSchemaException(message)

        self.assertEqual(str(ctx.exception), expected_message)

    def test_unsupported_field(self):
        message = "This is an internal assertion error."
        invalid_field = "testField"

        expected_message = (
            f"Unsupported JSON Schema: {message} (unsupported field: {invalid_field})"
        )

        with self.assertRaises(UnsupportedSchemaException) as ctx:
            raise UnsupportedSchemaException(message, unsupported_field=invalid_field)

        self.assertEqual(str(ctx.exception), expected_message)



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.type_parser_options import TypeParserOptions

from typing_extensions import (
    Iterable,
    Unpack,
)

import copy


class ArrayTypeParser(GenericTypeParser):
    mapped_type = list

    json_schema_type = "type:array"

    type_mappings = {
        "maxItems": "max_length",
        "minItems": "min_length",
    }

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        item_properties = kwargs.copy()
        item_properties["required"] = True

        if (items := properties.get("items")) is None:
            raise InvalidSchemaException(
                f"Array type {name} must have 'items' property defined.",
                invalid_field="items",
            )

        _item_type, _item_args = GenericTypeParser.type_from_properties(
            name, items, **item_properties
        )

        wrapper_type = set if properties.get("uniqueItems", False) else list
        field_type = wrapper_type[_item_type]

        mapped_properties = self.mappings_properties_builder(properties, **kwargs)

        if (
            default_value := mapped_properties.pop("default", None)
        ) is not None or not kwargs.get("required", False):
            mapped_properties["default_factory"] = self._build_default_factory(
                default_value, wrapper_type
            )

        if (example_values := mapped_properties.pop("examples", None)) is not None:
            mapped_properties["examples"] = [
                wrapper_type(example) for example in example_values
            ]

        return field_type, mapped_properties

    def _build_default_factory(self, default_list, wrapper_type):
        if default_list is None:
            return lambda: None

        if not isinstance(default_list, Iterable):
            raise InvalidSchemaException(
                f"Default value for array must be an iterable, got {type(default_list)}",
                invalid_field="default",
            )

        return lambda: copy.deepcopy(wrapper_type(default_list))



from jambo.parser import NullTypeParser

from unittest import TestCase


class TestNullTypeParser(TestCase):
    def test_null_parser_no_options(self):
        parser = NullTypeParser()

        properties = {"type": "null"}

        type_parsing, type_validator = parser.from_properties_impl(
            "placeholder", properties
        )

        self.assertEqual(type_parsing, type(None))
        self.assertEqual(type_validator, {"default": None})

    def test_null_parser_with_examples(self):
        parser = NullTypeParser()

        properties = {
            "type": "null",
            "examples": [None],
        }

        type_parsing, type_validator = parser.from_properties_impl(
            "placeholder", properties
        )

        self.assertEqual(type_parsing, type(None))
        self.assertEqual(type_validator["default"], None)
        self.assertEqual(type_validator["examples"], [None])

    def test_null_parser_with_invalid_default(self):
        parser = NullTypeParser()

        properties = {"type": "null", "default": "invalid"}

        type_parsing, type_validator = parser.from_properties_impl(
            "placeholder", properties
        )

        self.assertEqual(type_parsing, type(None))
        self.assertEqual(type_validator, {"default": None})



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.type_parser_options import TypeParserOptions

from pydantic import Field
from typing_extensions import Annotated, Union, Unpack


class AnyOfTypeParser(GenericTypeParser):
    mapped_type = Union

    json_schema_type = "anyOf"

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        if "anyOf" not in properties:
            raise InvalidSchemaException(
                f"AnyOf type {name} must have 'anyOf' property defined.",
                invalid_field="anyOf",
            )

        if not isinstance(properties["anyOf"], list):
            raise InvalidSchemaException(
                "AnyOf must be a list of types.", invalid_field="anyOf"
            )

        mapped_properties = self.mappings_properties_builder(properties, **kwargs)

        sub_properties = properties["anyOf"]

        sub_types = [
            GenericTypeParser.type_from_properties(
                f"{name}.sub{i}", subProperty, **kwargs
            )
            for i, subProperty in enumerate(sub_properties)
        ]

        if not kwargs.get("required", False):
            mapped_properties["default"] = mapped_properties.get("default")

        # By defining the type as Union of Annotated type we can use the Field validator
        # to enforce the constraints of each union type when needed.
        # We use Annotated to attach the Field validators to the type.
        field_types = []
        for subType, subProp in sub_types:
            default_value = subProp.pop("default", None)
            if default_value is None:
                default_value = ...

            field_types.append(Annotated[subType, Field(default_value, **subProp)])

        return Union[(*field_types,)], mapped_properties



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.type_parser_options import TypeParserOptions

from pydantic import AnyUrl, EmailStr
from typing_extensions import Any, Unpack

from datetime import date, datetime, time, timedelta
from ipaddress import IPv4Address, IPv6Address
from uuid import UUID


class StringTypeParser(GenericTypeParser):
    mapped_type = str

    json_schema_type = "type:string"

    type_mappings = {
        "maxLength": "max_length",
        "minLength": "min_length",
        "pattern": "pattern",
    }

    format_type_mapping = {
        # [7.3.1](https://json-schema.org/draft/2020-12/draft-bhutton-json-schema-validation-00#rfc.section.7.3.1). Dates, Times, and Duration
        "date": date,
        "time": time,
        "date-time": datetime,
        "duration": timedelta,
        # [7.3.2](https://json-schema.org/draft/2020-12/draft-bhutton-json-schema-validation-00#rfc.section.7.3.2). Email Addresses
        "email": EmailStr,
        # [7.3.3](https://json-schema.org/draft/2020-12/draft-bhutton-json-schema-validation-00#rfc.section.7.3.3). Hostnames
        "hostname": str,
        # [7.3.4](https://json-schema.org/draft/2020-12/draft-bhutton-json-schema-validation-00#rfc.section.7.3.4). IP Addresses
        "ipv4": IPv4Address,
        "ipv6": IPv6Address,
        # [7.3.5](https://json-schema.org/draft/2020-12/draft-bhutton-json-schema-validation-00#rfc.section.7.3.5). Resource Identifiers
        "uri": AnyUrl,
        # "iri" # Not supported by pydantic and currently not supported by jambo
        "uuid": UUID,
    }

    format_pattern_mapping = {
        "hostname": r"^[a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?(\.[a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?)*$",
    }

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        mapped_properties = self.mappings_properties_builder(properties, **kwargs)

        format_type = properties.get("format")
        if not format_type:
            return str, mapped_properties

        if format_type not in self.format_type_mapping:
            raise InvalidSchemaException(
                f"Unsupported string format: {format_type}", invalid_field="format"
            )

        mapped_type = self.format_type_mapping[format_type]
        if format_type in self.format_pattern_mapping:
            mapped_properties["pattern"] = self.format_pattern_mapping[format_type]

        if "examples" in mapped_properties:
            mapped_properties["examples"] = [
                self._parse_example(example, format_type, mapped_type)
                for example in mapped_properties["examples"]
            ]

        if "json_schema_extra" not in mapped_properties:
            mapped_properties["json_schema_extra"] = {}
        mapped_properties["json_schema_extra"]["format"] = format_type

        return mapped_type, mapped_properties

    def _parse_example(
        self, example: Any, format_type: str, mapped_type: type[Any]
    ) -> Any:
        """
        Parse example from JSON Schema format to python format
        :param example: Example Value
        :param format_type: Format Type
        :param mapped_type: Type to parse
        :return: Example parsed
        """
        match format_type:
            case "date" | "time" | "date-time":
                return mapped_type.fromisoformat(example)
            case "duration":
                # TODO: Implement duration parser
                raise NotImplementedError
            case "ipv4" | "ipv6":
                return mapped_type(example)
            case "uuid":
                return mapped_type(example)
            case _:
                return example



from __future__ import annotations

from typing_extensions import (
    Dict,
    List,
    Literal,
    TypedDict,
    Union,
)

from types import NoneType


# Primitive JSON types
JSONSchemaType = Literal[
    "string", "number", "integer", "boolean", "object", "array", "null"
]

JSONSchemaNativeTypes: tuple[type, ...] = (
    str,
    float,
    int,
    bool,
    list,
    set,
    NoneType,
)

JSONType = Union[str, int, float, bool, None, Dict[str, "JSONType"], List["JSONType"]]

# Dynamically define TypedDict with JSON Schema keywords
JSONSchema = TypedDict(
    "JSONSchema",
    {
        "$id": str,
        "$schema": str,
        "$ref": str,
        "$anchor": str,
        "$comment": str,
        "$defs": Dict[str, "JSONSchema"],
        "title": str,
        "description": str,
        "default": JSONType,
        "examples": List[JSONType],
        "type": JSONSchemaType | List[JSONSchemaType],
        "enum": List[JSONType],
        "const": JSONType,
        "properties": Dict[str, "JSONSchema"],
        "patternProperties": Dict[str, "JSONSchema"],
        "additionalProperties": Union[bool, "JSONSchema"],
        "required": List[str],
        "minProperties": int,
        "maxProperties": int,
        "dependencies": Dict[str, Union[List[str], "JSONSchema"]],
        "items": "JSONSchema",
        "prefixItems": List["JSONSchema"],
        "additionalItems": Union[bool, "JSONSchema"],
        "contains": "JSONSchema",
        "minItems": int,
        "maxItems": int,
        "uniqueItems": bool,
        "minLength": int,
        "maxLength": int,
        "pattern": str,
        "format": str,
        "minimum": float,
        "maximum": float,
        "exclusiveMinimum": Union[bool, float],
        "exclusiveMaximum": Union[bool, float],
        "multipleOf": float,
        "if": "JSONSchema",
        "then": "JSONSchema",
        "else": "JSONSchema",
        "allOf": List["JSONSchema"],
        "anyOf": List["JSONSchema"],
        "oneOf": List["JSONSchema"],
        "not": "JSONSchema",
    },
    total=False,  # all fields optional
)



from jambo.exceptions import InvalidSchemaException
from jambo.parser.anyof_type_parser import AnyOfTypeParser

from typing_extensions import Annotated, Union, get_args, get_origin

from unittest import TestCase


class TestAnyOfTypeParser(TestCase):
    def test_any_with_missing_properties(self):
        properties = {
            "notAnyOf": [
                {"type": "string"},
                {"type": "integer"},
            ],
        }

        with self.assertRaises(InvalidSchemaException):
            AnyOfTypeParser().from_properties("placeholder", properties)

    def test_any_of_with_invalid_properties(self):
        properties = {
            "anyOf": None,
        }

        with self.assertRaises(InvalidSchemaException):
            AnyOfTypeParser().from_properties("placeholder", properties)

    def test_any_of_string_or_int(self):
        """
        Tests the AnyOfTypeParser with a string or int type.
        """

        properties = {
            "anyOf": [
                {"type": "string"},
                {"type": "integer"},
            ],
        }

        type_parsing, _ = AnyOfTypeParser().from_properties(
            "placeholder", properties, required=True
        )

        # check union type has string and int
        self.assertEqual(get_origin(type_parsing), Union)

        type_1, type_2 = get_args(type_parsing)

        self.assertEqual(get_origin(type_1), Annotated)
        self.assertIn(str, get_args(type_1))

        self.assertEqual(get_origin(type_2), Annotated)
        self.assertIn(int, get_args(type_2))

    def test_any_of_string_or_int_with_default(self):
        """
        Tests the AnyOfTypeParser with a string or int type and a default value.
        """

        properties = {
            "anyOf": [
                {"type": "string"},
                {"type": "integer"},
            ],
            "default": 42,
        }

        type_parsing, type_validator = AnyOfTypeParser().from_properties(
            "placeholder", properties
        )

        # check union type has string and int
        self.assertEqual(get_origin(type_parsing), Union)

        type_1, type_2 = get_args(type_parsing)

        self.assertEqual(get_origin(type_1), Annotated)
        self.assertIn(str, get_args(type_1))

        self.assertEqual(get_origin(type_2), Annotated)
        self.assertIn(int, get_args(type_2))

        self.assertEqual(type_validator["default"], 42)

    def test_any_string_or_int_with_invalid_defaults(self):
        """
        Tests the AnyOfTypeParser with a string or int type and an invalid default value.
        """

        properties = {
            "anyOf": [
                {"type": "string"},
                {"type": "integer"},
            ],
            "default": 3.14,
        }

        with self.assertRaises(InvalidSchemaException):
            AnyOfTypeParser().from_properties("placeholder", properties)

    def test_anyof_with_examples(self):
        """
        Tests the AnyOfTypeParser with a string or int type and examples.
        """

        properties = {
            "anyOf": [
                {
                    "type": "string",
                    "examples": ["example string"],
                },
                {
                    "type": "integer",
                    "examples": [123],
                },
            ],
        }

        parsed_type, _ = AnyOfTypeParser().from_properties("placeholder", properties)

        type_1, type_2 = get_args(parsed_type)

        self.assertEqual(get_args(type_1)[1].examples, ["example string"])

        self.assertEqual(get_args(type_2)[1].examples, [123])

    def test_any_of_with_root_examples(self):
        """
        Tests the AnyOfTypeParser with a string or int type and examples.
        """

        properties = {
            "anyOf": [
                {"type": "string"},
                {"type": "integer"},
            ],
            "examples": ["100", 100],
        }

        _, type_validator = AnyOfTypeParser().from_properties("placeholder", properties)

        self.assertEqual(type_validator["examples"], ["100", 100])



from jambo.exceptions import InvalidSchemaException
from jambo.parser import IntTypeParser

from unittest import TestCase


class TestIntTypeParser(TestCase):
    def test_int_parser_no_options(self):
        parser = IntTypeParser()

        properties = {"type": "integer"}

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, int)
        self.assertEqual(type_validator, {"default": None})

    def test_int_parser_with_options(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "maximum": 10,
            "minimum": 1,
            "multipleOf": 2,
            "examples": [2, 4],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, int)
        self.assertEqual(type_validator["le"], 10)
        self.assertEqual(type_validator["ge"], 1)
        self.assertEqual(type_validator["multiple_of"], 2)
        self.assertEqual(type_validator["examples"], [2, 4])

    def test_int_parser_with_default(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "default": 6,
            "maximum": 10,
            "minimum": 1,
            "multipleOf": 2,
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing, int)
        self.assertEqual(type_validator["default"], 6)
        self.assertEqual(type_validator["le"], 10)
        self.assertEqual(type_validator["ge"], 1)
        self.assertEqual(type_validator["multiple_of"], 2)

    def test_int_parser_with_default_invalid_type(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "default": "invalid",  # Invalid default value
            "maximum": 10,
            "minimum": 1,
            "multipleOf": 2,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_int_parser_with_default_invalid_maximum(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "default": 15,
            "maximum": 10,
            "minimum": 1,
            "multipleOf": 2,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_int_parser_with_default_invalid_minimum(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "default": -5,
            "maximum": 10,
            "minimum": 1,
            "multipleOf": 2,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_int_parser_with_default_invalid_exclusive_maximum(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "default": 10,
            "exclusiveMaximum": 10,
            "minimum": 1,
            "multipleOf": 2,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_int_parser_with_default_invalid_exclusive_minimum(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "default": 1,
            "exclusiveMinimum": 1,
            "maximum": 10,
            "multipleOf": 2,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_int_parser_with_default_invalid_multipleOf(self):
        parser = IntTypeParser()

        properties = {
            "type": "integer",
            "default": 5,
            "maximum": 10,
            "minimum": 1,
            "multipleOf": 2,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)



from jambo.exceptions import InvalidSchemaException
from jambo.parser import StringTypeParser
from jambo.parser._type_parser import GenericTypeParser

from unittest import TestCase


class TestGenericTypeParser(TestCase):
    def test_get_impl(self):
        parser = GenericTypeParser._get_impl({"type": "string"})

        self.assertIsInstance(parser(), StringTypeParser)

    def test_get_impl_invalid_json_schema(self):
        with self.assertRaises(RuntimeError):
            StringTypeParser.json_schema_type = None
            GenericTypeParser._get_impl({"type": "string"})
        StringTypeParser.json_schema_type = "type:string"

    def test_get_impl_invalid_type(self):
        with self.assertRaises(InvalidSchemaException):
            GenericTypeParser._get_impl({"type": "invalid_type"})

    def test_invalid_examples_not_list(self):
        parser = StringTypeParser()

        properties = {
            "type": "integer",
            "examples": "this should be a list",
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)



from typing_extensions import Optional


class UnsupportedSchemaException(ValueError):
    """Exception raised for unsupported JSON schemas."""

    def __init__(
        self,
        message: str,
        unsupported_field: Optional[str] = None,
        cause: Optional[BaseException] = None,
    ) -> None:
        # Normalize message by stripping redundant prefix if present
        message = message.removeprefix("Unsupported JSON Schema: ")
        self.unsupported_field = unsupported_field
        self.cause = cause
        super().__init__(message)

    def __str__(self) -> str:
        base_msg = f"Unsupported JSON Schema: {super().__str__()}"
        if self.unsupported_field:
            return f"{base_msg} (unsupported field: {self.unsupported_field})"
        return base_msg



from jambo.exceptions import InvalidSchemaException
from jambo.parser import ArrayTypeParser

from typing_extensions import get_args

from unittest import TestCase


class TestArrayTypeParser(TestCase):
    def test_array_parser_no_options(self):
        parser = ArrayTypeParser()

        properties = {"items": {"type": "string"}}

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        element_type = get_args(type_parsing)[0]

        self.assertEqual(type_parsing.__origin__, list)
        self.assertEqual(element_type, str)

    def test_array_parser_with_no_items(self):
        parser = ArrayTypeParser()

        properties = {
            "default": ["a", "b", "c", "d"],
            "maxItems": 3,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_array_parser_with_options_unique(self):
        parser = ArrayTypeParser()

        properties = {"items": {"type": "string"}, "uniqueItems": True}

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing.__origin__, set)

    def test_array_parser_with_options_max_min(self):
        parser = ArrayTypeParser()

        properties = {"items": {"type": "string"}, "maxItems": 10, "minItems": 1}

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing.__origin__, list)
        self.assertEqual(type_validator["max_length"], 10)
        self.assertEqual(type_validator["min_length"], 1)

    def test_array_parser_with_options_default_list(self):
        parser = ArrayTypeParser()

        properties = {"items": {"type": "string"}, "default": ["a", "b", "c"]}

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing.__origin__, list)
        self.assertEqual(type_validator["default_factory"](), ["a", "b", "c"])

    def test_array_parse_with_options_default_set(self):
        parser = ArrayTypeParser()

        properties = {
            "items": {"type": "string"},
            "uniqueItems": True,
            "default": ["a", "b", "c"],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing.__origin__, set)
        self.assertEqual(type_validator["default_factory"](), {"a", "b", "c"})

    def test_array_parser_with_invalid_default_elem_type(self):
        parser = ArrayTypeParser()

        properties = {"items": {"type": "string"}, "default": ["a", 1, "c"]}

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_array_parser_with_invalid_default_type(self):
        parser = ArrayTypeParser()

        properties = {"items": {"type": "string"}, "default": 000}

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties=properties)

    def test_array_parser_with_invalid_default_min(self):
        parser = ArrayTypeParser()

        properties = {"items": {"type": "string"}, "default": ["a"], "minItems": 2}

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_array_parser_with_invalid_default_max(self):
        parser = ArrayTypeParser()

        properties = {
            "items": {"type": "string"},
            "default": ["a", "b", "c", "d"],
            "maxItems": 3,
        }

        with self.assertRaises(InvalidSchemaException):
            parser.from_properties("placeholder", properties)

    def test_array_parser_with_examples(self):
        parser = ArrayTypeParser()

        properties = {
            "items": {"type": "integer"},
            "examples": [
                [1, 2, 3],
                [4, 5, 6],
            ],
        }

        type_parsing, type_validator = parser.from_properties("placeholder", properties)

        self.assertEqual(type_parsing.__origin__, list)
        self.assertEqual(type_validator["examples"], [[1, 2, 3], [4, 5, 6]])



from jambo.exceptions import InternalAssertionException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.json_schema_type import JSONSchema
from jambo.types.type_parser_options import TypeParserOptions

from pydantic import BaseModel, ConfigDict, Field, create_model
from pydantic.fields import FieldInfo
from typing_extensions import Unpack

import warnings


class ObjectTypeParser(GenericTypeParser):
    mapped_type = object

    json_schema_type = "type:object"

    def from_properties_impl(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ) -> tuple[type[BaseModel], dict]:
        type_parsing = self.to_model(
            name,
            properties.get("properties", {}),
            properties.get("required", []),
            **kwargs,
        )
        type_properties = self.mappings_properties_builder(properties, **kwargs)

        if (
            default_value := type_properties.pop("default", None)
        ) is not None or not kwargs.get("required", False):
            type_properties["default_factory"] = (
                lambda: type_parsing.model_validate(default_value)
                if default_value is not None
                else None
            )

        if (example_values := type_properties.pop("examples", None)) is not None:
            type_properties["examples"] = [
                type_parsing.model_validate(example) for example in example_values
            ]

        return type_parsing, type_properties

    @classmethod
    def to_model(
        cls,
        name: str,
        properties: dict[str, JSONSchema],
        required_keys: list[str],
        **kwargs: Unpack[TypeParserOptions],
    ) -> type[BaseModel]:
        """
        Converts JSON Schema object properties to a Pydantic model.
        :param name: The name of the model.
        :param properties: The properties of the JSON Schema object.
        :param required_keys: List of required keys in the schema.
        :return: A Pydantic model class.
        """
        ref_cache = kwargs.get("ref_cache")
        if ref_cache is None:
            raise InternalAssertionException(
                "`ref_cache` must be provided in kwargs for ObjectTypeParser"
            )

        if (model := ref_cache.get(name)) is not None and isinstance(model, type):
            warnings.warn(
                f"Type '{name}' is already in the ref_cache and therefore cached value will be used."
                " This may indicate a namming collision in the schema or just a normal optimization,"
                " if this behavior is desired pass a clean ref_cache or use the param `without_cache`"
            )
            return model

        model_config = ConfigDict(validate_assignment=True)
        fields = cls._parse_properties(name, properties, required_keys, **kwargs)

        model = create_model(name, __config__=model_config, **fields)  # type: ignore
        ref_cache[name] = model

        return model

    @classmethod
    def _parse_properties(
        cls,
        name: str,
        properties: dict[str, JSONSchema],
        required_keys: list[str],
        **kwargs: Unpack[TypeParserOptions],
    ) -> dict[str, tuple[type, FieldInfo]]:
        required_keys = required_keys or []

        fields = {}
        for field_name, field_prop in properties.items():
            sub_property: TypeParserOptions = kwargs.copy()
            sub_property["required"] = field_name in required_keys

            parsed_type, parsed_properties = GenericTypeParser.type_from_properties(
                f"{name}.{field_name}",
                field_prop,
                **sub_property,  # type: ignore
            )
            fields[field_name] = (parsed_type, Field(**parsed_properties))

        return fields



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.json_schema_type import JSONSchemaNativeTypes
from jambo.types.type_parser_options import JSONSchema, TypeParserOptions

from typing_extensions import Unpack

from enum import Enum


class EnumTypeParser(GenericTypeParser):
    json_schema_type = "enum"

    def from_properties_impl(
        self, name: str, properties: JSONSchema, **kwargs: Unpack[TypeParserOptions]
    ):
        if "enum" not in properties:
            raise InvalidSchemaException(
                f"Enum type {name} must have 'enum' property defined.",
                invalid_field="enum",
            )

        enum_values = properties["enum"]

        if not isinstance(enum_values, list):
            raise InvalidSchemaException(
                f"Enum type {name} must have 'enum' as a list of values.",
                invalid_field="enum",
            )

        if any(not isinstance(value, JSONSchemaNativeTypes) for value in enum_values):
            raise InvalidSchemaException(
                f"Enum type {name} must have 'enum' values of allowed types: {JSONSchemaNativeTypes}.",
                invalid_field="enum",
            )

        # Create a new Enum type dynamically
        enum_type = Enum(name, {str(value).upper(): value for value in enum_values})  # type: ignore
        parsed_properties = self.mappings_properties_builder(properties, **kwargs)

        if "default" in parsed_properties and parsed_properties["default"] is not None:
            parsed_properties["default"] = enum_type(parsed_properties["default"])

        if "examples" in parsed_properties:
            parsed_properties["examples"] = [
                enum_type(example) for example in parsed_properties["examples"]
            ]

        return enum_type, parsed_properties



from jambo.exceptions import InvalidSchemaException
from jambo.parser.allof_type_parser import AllOfTypeParser

from pydantic import ValidationError

from unittest import TestCase


class TestAllOfTypeParser(TestCase):
    def test_all_of_type_parser_object_type(self):
        """
        Test the AllOfTypeParser with an object type and validate the properties.
        When using allOf with object it should be able to validate the properties
        and join them correctly.
        """
        properties = {
            "type": "object",
            "allOf": [
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "minLength": 1,
                        }
                    },
                },
                {
                    "type": "object",
                    "properties": {
                        "name": {
                            "type": "string",
                            "maxLength": 4,
                        },
                        "age": {
                            "type": "integer",
                            "maximum": 100,
                            "minimum": 0,
                        },
                    },
                },
            ],
        }

        type_parsing, type_validator = AllOfTypeParser().from_properties(
            "placeholder", properties, ref_cache={}
        )

        with self.assertRaises(ValidationError):
            type_parsing(name="John", age=101)

        with self.assertRaises(ValidationError):
            type_parsing(name="", age=30)

        with self.assertRaises(ValidationError):
            type_parsing(name="John Invalid", age=30)

        obj = type_parsing(name="John", age=30)
        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)

    def test_all_of_type_parser_object_type_required(self):
        """
        Tests the required properties of the AllOfTypeParser with an object type.
        """

        properties = {
            "type": "object",
            "allOf": [
                {
                    "properties": {
                        "name": {
                            "type": "string",
                        }
                    },
                    "required": ["name"],
                },
                {
                    "type": "object",
                    "properties": {
                        "age": {
                            "type": "integer",
                        }
                    },
                    "required": ["age"],
                },
            ],
        }

        type_parsing, type_validator = AllOfTypeParser().from_properties(
            "placeholder", properties, ref_cache={}
        )

        with self.assertRaises(ValidationError):
            type_parsing(name="John")

        with self.assertRaises(ValidationError):
            type_parsing(age=30)

        obj = type_parsing(name="John", age=30)
        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)

    def test_all_of_type_top_level_type(self):
        """
        Tests the AllOfTypeParser with a top-level type and validate the properties.
        """

        properties = {
            "type": "string",
            "allOf": [
                {"maxLength": 11},
                {"maxLength": 4},
                {"minLength": 1},
                {"minLength": 2},
            ],
        }

        type_parsing, type_validator = AllOfTypeParser().from_properties(
            "placeholder", properties, ref_cache={}
        )

        self.assertEqual(type_parsing, str)
        self.assertEqual(type_validator["max_length"], 11)
        self.assertEqual(type_validator["min_length"], 1)

    def test_all_of_type_parser_in_fields(self):
        """
        Tests the AllOfTypeParser when set in the fields of a model.
        """
        properties = {
            "allOf": [
                {"type": "string", "maxLength": 11},
                {"type": "string", "maxLength": 4},
                {"type": "string", "minLength": 1},
                {"type": "string", "minLength": 2},
            ]
        }

        type_parsing, type_validator = AllOfTypeParser().from_properties(
            "placeholder", properties, ref_cache={}
        )

        self.assertEqual(type_parsing, str)
        self.assertEqual(type_validator["max_length"], 11)
        self.assertEqual(type_validator["min_length"], 1)

    def test_invalid_all_of(self):
        """
        Tests that an error is raised when the allOf type is not present.
        """
        properties = {
            "wrongKey": [
                {"type": "string", "maxLength": 11},
                {"type": "string", "maxLength": 4},
                {"type": "string", "minLength": 1},
                {"type": "string", "minLength": 2},
            ]
        }

        with self.assertRaises(InvalidSchemaException):
            AllOfTypeParser().from_properties("placeholder", properties, ref_cache={})

    def test_all_of_invalid_type_not_present(self):
        properties = {
            "allOf": [
                {"maxLength": 11},
                {"maxLength": 4},
                {"minLength": 1},
                {"minLength": 2},
            ]
        }

        with self.assertRaises(InvalidSchemaException):
            AllOfTypeParser().from_properties("placeholder", properties, ref_cache={})

    def test_all_of_invalid_type_in_fields(self):
        properties = {
            "allOf": [
                {"type": "string", "maxLength": 11},
                {"type": "integer", "maxLength": 4},
                {"type": "string", "minLength": 1},
                {"minLength": 2},
            ]
        }

        with self.assertRaises(InvalidSchemaException):
            AllOfTypeParser().from_properties("placeholder", properties, ref_cache={})

    def test_all_of_invalid_type_not_all_equal(self):
        """
        Tests that an error is raised when the allOf types are not all equal.
        """

        properties = {
            "allOf": [
                {"type": "string", "maxLength": 11},
                {"type": "integer", "maxLength": 4},
                {"type": "string", "minLength": 1},
            ]
        }

        with self.assertRaises(InvalidSchemaException):
            AllOfTypeParser().from_properties("placeholder", properties, ref_cache={})

    def test_all_of_description_field(self):
        """
        Tests the AllOfTypeParser with a description field.
        """

        properties = {
            "type": "object",
            "allOf": [
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "description": "One",
                        }
                    },
                },
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "description": "Of",
                        }
                    },
                },
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "description": "Us",
                        }
                    },
                },
            ],
        }

        type_parsing, _ = AllOfTypeParser().from_properties(
            "placeholder", properties, ref_cache={}
        )

        self.assertEqual(
            type_parsing.model_json_schema()["properties"]["name"]["description"],
            "One | Of | Us",
        )

    def test_all_of_with_defaults(self):
        """
        Tests the AllOfTypeParser with a default value.
        """

        properties = {
            "type": "object",
            "allOf": [
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "default": "John",
                        }
                    },
                },
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "default": "John",
                        },
                        "age": {
                            "type": "integer",
                            "default": 30,
                        },
                    },
                },
            ],
        }

        type_parsing, _ = AllOfTypeParser().from_properties(
            "placeholder", properties, ref_cache={}
        )
        obj = type_parsing()
        self.assertEqual(obj.name, "John")
        self.assertEqual(obj.age, 30)

    def test_all_of_with_conflicting_defaults(self):
        """
        Tests the AllOfTypeParser with conflicting default values.
        """

        properties = {
            "type": "object",
            "allOf": [
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "default": "John",
                        }
                    },
                },
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "default": "Doe",
                        }
                    },
                },
            ],
        }

        with self.assertRaises(InvalidSchemaException):
            AllOfTypeParser().from_properties("placeholder", properties, ref_cache={})

    def test_all_of_with_root_examples(self):
        """
        Tests the AllOfTypeParser with examples.
        """

        properties = {
            "type": "object",
            "allOf": [
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "minLength": 1,
                        }
                    },
                },
                {
                    "properties": {
                        "name": {
                            "type": "string",
                            "maxLength": 4,
                        }
                    },
                },
            ],
            "examples": [
                {"name": "John"},
                {"name": "Jane"},
                {"name": "Doe"},
                {"name": "Jack"},
            ],
        }

        type_parsed, type_properties = AllOfTypeParser().from_properties(
            "placeholder", properties, ref_cache={}
        )

        self.assertEqual(
            type_properties["examples"],
            [
                type_parsed(name="John"),
                type_parsed(name="Jane"),
                type_parsed(name="Doe"),
                type_parsed(name="Jack"),
            ],
        )



from jambo.exceptions import InvalidSchemaException
from jambo.parser._type_parser import GenericTypeParser
from jambo.types.json_schema_type import JSONSchemaNativeTypes
from jambo.types.type_parser_options import TypeParserOptions

from pydantic import AfterValidator
from typing_extensions import Annotated, Any, Literal, Unpack


class ConstTypeParser(GenericTypeParser):
    json_schema_type = "const"

    default_mappings = {
        "const": "default",
        "description": "description",
        "examples": "examples",
    }

    def from_properties_impl(
        self, name, properties, **kwargs: Unpack[TypeParserOptions]
    ):
        if "const" not in properties:
            raise InvalidSchemaException(
                f"Const type {name} must have 'const' property defined.",
                invalid_field="const",
            )

        const_value = properties["const"]

        if not isinstance(const_value, JSONSchemaNativeTypes):
            raise InvalidSchemaException(
                f"Const type {name} must have 'const' value of allowed types: {JSONSchemaNativeTypes}.",
                invalid_field="const",
            )

        const_type = self._build_const_type(const_value)
        parsed_properties = self.mappings_properties_builder(properties, **kwargs)

        return const_type, parsed_properties

    def _build_const_type(self, const_value):
        # Try to use Literal for hashable types (required for discriminated unions)
        # Fall back to validator approach for non-hashable types
        try:
            # Test if the value is hashable (can be used in Literal)
            hash(const_value)
            return Literal[const_value]
        except TypeError:
            # Non-hashable type (like list, dict), use validator approach
            def _validate_const_value(value: Any) -> Any:
                if value != const_value:
                    raise ValueError(
                        f"Value must be equal to the constant value: {const_value}"
                    )
                return value

            return Annotated[type(const_value), AfterValidator(_validate_const_value)]



from typing_extensions import Optional


class InvalidSchemaException(ValueError):
    """Exception raised for invalid JSON schemas."""

    def __init__(
        self,
        message: str,
        invalid_field: Optional[str] = None,
        cause: Optional[BaseException] = None,
    ) -> None:
        # Normalize message by stripping redundant prefix if present
        message = message.removeprefix("Invalid JSON Schema: ")
        self.invalid_field = invalid_field
        self.cause = cause
        super().__init__(message)

    def __str__(self) -> str:
        base_msg = f"Invalid JSON Schema: {super().__str__()}"
        if self.invalid_field:
            return f"{base_msg} (invalid field: {self.invalid_field})"
        if self.cause:
            return (
                f"{base_msg} (caused by {self.cause.__class__.__name__}: {self.cause})"
            )
        return base_msg
